from .base_classes import *
from .support_functions import *


class Provenance(Resource):
    """Provenance information that describes the activity that led to the
    creation of a set of resources. This information can be used to
    help determine their reliability or trace where the information
    in them came from. The focus of the provenance resource is
    record keeping, audit and traceability, and not explicit
    statements of clinical significance.If the element is present,
    it must have either a @value, an @id, or extensions"""
    subclass = None
    superclass = Resource
    def __init__(self, id=None, extension=None, modifierExtension=None, language=None, text=None, contained=None, target=None, period=None, recorded=None, reason=None, location=None, policy=None, agent=None, entity=None, integritySignature=None):
        self.original_tagname_ = None
        super(Provenance, self).__init__(id, extension, modifierExtension, language, text, contained, )
        if target is None:
            self.target = []
        else:
            self.target = target
        self.period = period
        self.recorded = recorded
        self.reason = reason
        self.location = location
        if policy is None:
            self.policy = []
        else:
            self.policy = policy
        if agent is None:
            self.agent = []
        else:
            self.agent = agent
        if entity is None:
            self.entity = []
        else:
            self.entity = entity
        self.integritySignature = integritySignature
    def factory(*args_, **kwargs_):
        if Provenance.subclass:
            return Provenance.subclass(*args_, **kwargs_)
        else:
            return Provenance(*args_, **kwargs_)
    factory = staticmethod(factory)
    def get_target(self): return self.target
    def set_target(self, target): self.target = target
    def add_target(self, value): self.target.append(value)
    def insert_target(self, index, value): self.target[index] = value
    def get_period(self): return self.period
    def set_period(self, period): self.period = period
    def get_recorded(self): return self.recorded
    def set_recorded(self, recorded): self.recorded = recorded
    def get_reason(self): return self.reason
    def set_reason(self, reason): self.reason = reason
    def get_location(self): return self.location
    def set_location(self, location): self.location = location
    def get_policy(self): return self.policy
    def set_policy(self, policy): self.policy = policy
    def add_policy(self, value): self.policy.append(value)
    def insert_policy(self, index, value): self.policy[index] = value
    def get_agent(self): return self.agent
    def set_agent(self, agent): self.agent = agent
    def add_agent(self, value): self.agent.append(value)
    def insert_agent(self, index, value): self.agent[index] = value
    def get_entity(self): return self.entity
    def set_entity(self, entity): self.entity = entity
    def add_entity(self, value): self.entity.append(value)
    def insert_entity(self, index, value): self.entity[index] = value
    def get_integritySignature(self): return self.integritySignature
    def set_integritySignature(self, integritySignature): self.integritySignature = integritySignature
    def hasContent_(self):
        if (
            self.target or
            self.period is not None or
            self.recorded is not None or
            self.reason is not None or
            self.location is not None or
            self.policy or
            self.agent or
            self.entity or
            self.integritySignature is not None or
            super(Provenance, self).hasContent_()
        ):
            return True
        else:
            return False
    def export(self, outfile, level, namespace_='', name_='Provenance', namespacedef_='', pretty_print=True):
        if pretty_print:
            eol_ = '\n'
        else:
            eol_ = ''
        if self.original_tagname_ is not None:
            name_ = self.original_tagname_
        showIndent(outfile, level, pretty_print)
        outfile.write('<%s%s%s' % (namespace_, name_, namespacedef_ and ' ' + namespacedef_ or '', ))
        already_processed = set()
        self.exportAttributes(outfile, level, already_processed, namespace_, name_='Provenance')
        if self.hasContent_():
            outfile.write('>%s' % (eol_, ))
            self.exportChildren(outfile, level + 1, namespace_='', name_='Provenance', pretty_print=pretty_print)
            showIndent(outfile, level, pretty_print)
            outfile.write('</%s%s>%s' % (namespace_, name_, eol_))
        else:
            outfile.write('/>%s' % (eol_, ))
    def exportAttributes(self, outfile, level, already_processed, namespace_='', name_='Provenance'):
        super(Provenance, self).exportAttributes(outfile, level, already_processed, namespace_, name_='Provenance')
    def exportChildren(self, outfile, level, namespace_='', name_='Provenance', fromsubclass_=False, pretty_print=True):
        super(Provenance, self).exportChildren(outfile, level, namespace_, name_, True, pretty_print=pretty_print)
        if pretty_print:
            eol_ = '\n'
        else:
            eol_ = ''
        for target_ in self.target:
            target_.export(outfile, level, namespace_, name_='target', pretty_print=pretty_print)
        if self.period is not None:
            self.period.export(outfile, level, namespace_, name_='period', pretty_print=pretty_print)
        if self.recorded is not None:
            self.recorded.export(outfile, level, namespace_, name_='recorded', pretty_print=pretty_print)
        if self.reason is not None:
            self.reason.export(outfile, level, namespace_, name_='reason', pretty_print=pretty_print)
        if self.location is not None:
            self.location.export(outfile, level, namespace_, name_='location', pretty_print=pretty_print)
        for policy_ in self.policy:
            policy_.export(outfile, level, namespace_, name_='policy', pretty_print=pretty_print)
        for agent_ in self.agent:
            agent_.export(outfile, level, namespace_, name_='agent', pretty_print=pretty_print)
        for entity_ in self.entity:
            entity_.export(outfile, level, namespace_, name_='entity', pretty_print=pretty_print)
        if self.integritySignature is not None:
            self.integritySignature.export(outfile, level, namespace_, name_='integritySignature', pretty_print=pretty_print)
    def to_etree(self, parent_element=None, name_='Provenance', mapping_=None):
        element = super(Provenance, self).to_etree(parent_element, name_, mapping_)
        for target_ in self.target:
            target_.to_etree(element, name_='target', mapping_=mapping_)
        if self.period is not None:
            period_ = self.period
            period_.to_etree(element, name_='period', mapping_=mapping_)
        if self.recorded is not None:
            recorded_ = self.recorded
            recorded_.to_etree(element, name_='recorded', mapping_=mapping_)
        if self.reason is not None:
            reason_ = self.reason
            reason_.to_etree(element, name_='reason', mapping_=mapping_)
        if self.location is not None:
            location_ = self.location
            location_.to_etree(element, name_='location', mapping_=mapping_)
        for policy_ in self.policy:
            policy_.to_etree(element, name_='policy', mapping_=mapping_)
        for agent_ in self.agent:
            agent_.to_etree(element, name_='agent', mapping_=mapping_)
        for entity_ in self.entity:
            entity_.to_etree(element, name_='entity', mapping_=mapping_)
        if self.integritySignature is not None:
            integritySignature_ = self.integritySignature
            integritySignature_.to_etree(element, name_='integritySignature', mapping_=mapping_)
        if mapping_ is not None:
            mapping_[self] = element
        return element
    def build(self, node):
        already_processed = set()
        self.buildAttributes(node, node.attrib, already_processed)
        for child in node:
            nodeName_ = Tag_pattern_.match(child.tag).groups()[-1]
            self.buildChildren(child, node, nodeName_)
        return self
    def buildAttributes(self, node, attrs, already_processed):
        super(Provenance, self).buildAttributes(node, attrs, already_processed)
    def buildChildren(self, child_, node, nodeName_, fromsubclass_=False):
        if nodeName_ == 'target':
            obj_ = ResourceReference.factory()
            obj_.build(child_)
            self.target.append(obj_)
            obj_.original_tagname_ = 'target'
        elif nodeName_ == 'period':
            obj_ = Period.factory()
            obj_.build(child_)
            self.period = obj_
            obj_.original_tagname_ = 'period'
        elif nodeName_ == 'recorded':
            obj_ = instant.factory()
            obj_.build(child_)
            self.recorded = obj_
            obj_.original_tagname_ = 'recorded'
        elif nodeName_ == 'reason':
            obj_ = CodeableConcept.factory()
            obj_.build(child_)
            self.reason = obj_
            obj_.original_tagname_ = 'reason'
        elif nodeName_ == 'location':
            obj_ = ResourceReference.factory()
            obj_.build(child_)
            self.location = obj_
            obj_.original_tagname_ = 'location'
        elif nodeName_ == 'policy':
            obj_ = uri.factory()
            obj_.build(child_)
            self.policy.append(obj_)
            obj_.original_tagname_ = 'policy'
        elif nodeName_ == 'agent':
            obj_ = Provenance_Agent.factory()
            obj_.build(child_)
            self.agent.append(obj_)
            obj_.original_tagname_ = 'agent'
        elif nodeName_ == 'entity':
            obj_ = Provenance_Entity.factory()
            obj_.build(child_)
            self.entity.append(obj_)
            obj_.original_tagname_ = 'entity'
        elif nodeName_ == 'integritySignature':
            obj_ = string.factory()
            obj_.build(child_)
            self.integritySignature = obj_
            obj_.original_tagname_ = 'integritySignature'
        super(Provenance, self).buildChildren(child_, node, nodeName_, True)
# end class Provenance


class Provenance_Agent(BackboneElement):
    """Provenance information that describes the activity that led to the
    creation of a set of resources. This information can be used to
    help determine their reliability or trace where the information
    in them came from. The focus of the provenance resource is
    record keeping, audit and traceability, and not explicit
    statements of clinical significance."""
    subclass = None
    superclass = BackboneElement
    def __init__(self, id=None, extension=None, modifierExtension=None, role=None, type_=None, reference=None, display=None):
        self.original_tagname_ = None
        super(Provenance_Agent, self).__init__(id, extension, modifierExtension, )
        self.role = role
        self.type_ = type_
        self.reference = reference
        self.display = display
    def factory(*args_, **kwargs_):
        if Provenance_Agent.subclass:
            return Provenance_Agent.subclass(*args_, **kwargs_)
        else:
            return Provenance_Agent(*args_, **kwargs_)
    factory = staticmethod(factory)
    def get_role(self): return self.role
    def set_role(self, role): self.role = role
    def get_type(self): return self.type_
    def set_type(self, type_): self.type_ = type_
    def get_reference(self): return self.reference
    def set_reference(self, reference): self.reference = reference
    def get_display(self): return self.display
    def set_display(self, display): self.display = display
    def hasContent_(self):
        if (
            self.role is not None or
            self.type_ is not None or
            self.reference is not None or
            self.display is not None or
            super(Provenance_Agent, self).hasContent_()
        ):
            return True
        else:
            return False
    def export(self, outfile, level, namespace_='', name_='Provenance.Agent', namespacedef_='', pretty_print=True):
        if pretty_print:
            eol_ = '\n'
        else:
            eol_ = ''
        if self.original_tagname_ is not None:
            name_ = self.original_tagname_
        showIndent(outfile, level, pretty_print)
        outfile.write('<%s%s%s' % (namespace_, name_, namespacedef_ and ' ' + namespacedef_ or '', ))
        already_processed = set()
        self.exportAttributes(outfile, level, already_processed, namespace_, name_='Provenance.Agent')
        if self.hasContent_():
            outfile.write('>%s' % (eol_, ))
            self.exportChildren(outfile, level + 1, namespace_='', name_='Provenance.Agent', pretty_print=pretty_print)
            showIndent(outfile, level, pretty_print)
            outfile.write('</%s%s>%s' % (namespace_, name_, eol_))
        else:
            outfile.write('/>%s' % (eol_, ))
    def exportAttributes(self, outfile, level, already_processed, namespace_='', name_='Provenance.Agent'):
        super(Provenance_Agent, self).exportAttributes(outfile, level, already_processed, namespace_, name_='Provenance.Agent')
    def exportChildren(self, outfile, level, namespace_='', name_='Provenance.Agent', fromsubclass_=False, pretty_print=True):
        super(Provenance_Agent, self).exportChildren(outfile, level, namespace_, name_, True, pretty_print=pretty_print)
        if pretty_print:
            eol_ = '\n'
        else:
            eol_ = ''
        if self.role is not None:
            self.role.export(outfile, level, namespace_, name_='role', pretty_print=pretty_print)
        if self.type_ is not None:
            self.type_.export(outfile, level, namespace_, name_='type', pretty_print=pretty_print)
        if self.reference is not None:
            self.reference.export(outfile, level, namespace_, name_='reference', pretty_print=pretty_print)
        if self.display is not None:
            self.display.export(outfile, level, namespace_, name_='display', pretty_print=pretty_print)
    def to_etree(self, parent_element=None, name_='Provenance.Agent', mapping_=None):
        element = super(Provenance_Agent, self).to_etree(parent_element, name_, mapping_)
        if self.role is not None:
            role_ = self.role
            role_.to_etree(element, name_='role', mapping_=mapping_)
        if self.type_ is not None:
            type__ = self.type_
            type__.to_etree(element, name_='type', mapping_=mapping_)
        if self.reference is not None:
            reference_ = self.reference
            reference_.to_etree(element, name_='reference', mapping_=mapping_)
        if self.display is not None:
            display_ = self.display
            display_.to_etree(element, name_='display', mapping_=mapping_)
        if mapping_ is not None:
            mapping_[self] = element
        return element
    def build(self, node):
        already_processed = set()
        self.buildAttributes(node, node.attrib, already_processed)
        for child in node:
            nodeName_ = Tag_pattern_.match(child.tag).groups()[-1]
            self.buildChildren(child, node, nodeName_)
        return self
    def buildAttributes(self, node, attrs, already_processed):
        super(Provenance_Agent, self).buildAttributes(node, attrs, already_processed)
    def buildChildren(self, child_, node, nodeName_, fromsubclass_=False):
        if nodeName_ == 'role':
            obj_ = Coding.factory()
            obj_.build(child_)
            self.role = obj_
            obj_.original_tagname_ = 'role'
        elif nodeName_ == 'type':
            obj_ = Coding.factory()
            obj_.build(child_)
            self.type_ = obj_
            obj_.original_tagname_ = 'type'
        elif nodeName_ == 'reference':
            obj_ = uri.factory()
            obj_.build(child_)
            self.reference = obj_
            obj_.original_tagname_ = 'reference'
        elif nodeName_ == 'display':
            obj_ = string.factory()
            obj_.build(child_)
            self.display = obj_
            obj_.original_tagname_ = 'display'
        super(Provenance_Agent, self).buildChildren(child_, node, nodeName_, True)
# end class Provenance_Agent


class Provenance_Entity(BackboneElement):
    """Provenance information that describes the activity that led to the
    creation of a set of resources. This information can be used to
    help determine their reliability or trace where the information
    in them came from. The focus of the provenance resource is
    record keeping, audit and traceability, and not explicit
    statements of clinical significance."""
    subclass = None
    superclass = BackboneElement
    def __init__(self, id=None, extension=None, modifierExtension=None, role=None, type_=None, reference=None, display=None, agent=None):
        self.original_tagname_ = None
        super(Provenance_Entity, self).__init__(id, extension, modifierExtension, )
        self.role = role
        self.type_ = type_
        self.reference = reference
        self.display = display
        self.agent = agent
    def factory(*args_, **kwargs_):
        if Provenance_Entity.subclass:
            return Provenance_Entity.subclass(*args_, **kwargs_)
        else:
            return Provenance_Entity(*args_, **kwargs_)
    factory = staticmethod(factory)
    def get_role(self): return self.role
    def set_role(self, role): self.role = role
    def get_type(self): return self.type_
    def set_type(self, type_): self.type_ = type_
    def get_reference(self): return self.reference
    def set_reference(self, reference): self.reference = reference
    def get_display(self): return self.display
    def set_display(self, display): self.display = display
    def get_agent(self): return self.agent
    def set_agent(self, agent): self.agent = agent
    def hasContent_(self):
        if (
            self.role is not None or
            self.type_ is not None or
            self.reference is not None or
            self.display is not None or
            self.agent is not None or
            super(Provenance_Entity, self).hasContent_()
        ):
            return True
        else:
            return False
    def export(self, outfile, level, namespace_='', name_='Provenance.Entity', namespacedef_='', pretty_print=True):
        if pretty_print:
            eol_ = '\n'
        else:
            eol_ = ''
        if self.original_tagname_ is not None:
            name_ = self.original_tagname_
        showIndent(outfile, level, pretty_print)
        outfile.write('<%s%s%s' % (namespace_, name_, namespacedef_ and ' ' + namespacedef_ or '', ))
        already_processed = set()
        self.exportAttributes(outfile, level, already_processed, namespace_, name_='Provenance.Entity')
        if self.hasContent_():
            outfile.write('>%s' % (eol_, ))
            self.exportChildren(outfile, level + 1, namespace_='', name_='Provenance.Entity', pretty_print=pretty_print)
            showIndent(outfile, level, pretty_print)
            outfile.write('</%s%s>%s' % (namespace_, name_, eol_))
        else:
            outfile.write('/>%s' % (eol_, ))
    def exportAttributes(self, outfile, level, already_processed, namespace_='', name_='Provenance.Entity'):
        super(Provenance_Entity, self).exportAttributes(outfile, level, already_processed, namespace_, name_='Provenance.Entity')
    def exportChildren(self, outfile, level, namespace_='', name_='Provenance.Entity', fromsubclass_=False, pretty_print=True):
        super(Provenance_Entity, self).exportChildren(outfile, level, namespace_, name_, True, pretty_print=pretty_print)
        if pretty_print:
            eol_ = '\n'
        else:
            eol_ = ''
        if self.role is not None:
            self.role.export(outfile, level, namespace_, name_='role', pretty_print=pretty_print)
        if self.type_ is not None:
            self.type_.export(outfile, level, namespace_, name_='type', pretty_print=pretty_print)
        if self.reference is not None:
            self.reference.export(outfile, level, namespace_, name_='reference', pretty_print=pretty_print)
        if self.display is not None:
            self.display.export(outfile, level, namespace_, name_='display', pretty_print=pretty_print)
        if self.agent is not None:
            self.agent.export(outfile, level, namespace_, name_='agent', pretty_print=pretty_print)
    def to_etree(self, parent_element=None, name_='Provenance.Entity', mapping_=None):
        element = super(Provenance_Entity, self).to_etree(parent_element, name_, mapping_)
        if self.role is not None:
            role_ = self.role
            role_.to_etree(element, name_='role', mapping_=mapping_)
        if self.type_ is not None:
            type__ = self.type_
            type__.to_etree(element, name_='type', mapping_=mapping_)
        if self.reference is not None:
            reference_ = self.reference
            reference_.to_etree(element, name_='reference', mapping_=mapping_)
        if self.display is not None:
            display_ = self.display
            display_.to_etree(element, name_='display', mapping_=mapping_)
        if self.agent is not None:
            agent_ = self.agent
            agent_.to_etree(element, name_='agent', mapping_=mapping_)
        if mapping_ is not None:
            mapping_[self] = element
        return element
    def build(self, node):
        already_processed = set()
        self.buildAttributes(node, node.attrib, already_processed)
        for child in node:
            nodeName_ = Tag_pattern_.match(child.tag).groups()[-1]
            self.buildChildren(child, node, nodeName_)
        return self
    def buildAttributes(self, node, attrs, already_processed):
        super(Provenance_Entity, self).buildAttributes(node, attrs, already_processed)
    def buildChildren(self, child_, node, nodeName_, fromsubclass_=False):
        if nodeName_ == 'role':
            obj_ = ProvenanceEntityRole.factory()
            obj_.build(child_)
            self.role = obj_
            obj_.original_tagname_ = 'role'
        elif nodeName_ == 'type':
            obj_ = Coding.factory()
            obj_.build(child_)
            self.type_ = obj_
            obj_.original_tagname_ = 'type'
        elif nodeName_ == 'reference':
            obj_ = uri.factory()
            obj_.build(child_)
            self.reference = obj_
            obj_.original_tagname_ = 'reference'
        elif nodeName_ == 'display':
            obj_ = string.factory()
            obj_.build(child_)
            self.display = obj_
            obj_.original_tagname_ = 'display'
        elif nodeName_ == 'agent':
            obj_ = Provenance_Agent.factory()
            obj_.build(child_)
            self.agent = obj_
            obj_.original_tagname_ = 'agent'
        super(Provenance_Entity, self).buildChildren(child_, node, nodeName_, True)
# end class Provenance_Entity


class ProvenanceEntityRole(Element):
    """How an entity was used in an activityIf the element is present, it
    must have either a @value, an @id, or extensions"""
    subclass = None
    superclass = Element
    def __init__(self, id=None, extension=None, value=None):
        self.original_tagname_ = None
        super(ProvenanceEntityRole, self).__init__(id, extension, )
        self.value = _cast(None, value)
    def factory(*args_, **kwargs_):
        if ProvenanceEntityRole.subclass:
            return ProvenanceEntityRole.subclass(*args_, **kwargs_)
        else:
            return ProvenanceEntityRole(*args_, **kwargs_)
    factory = staticmethod(factory)
    def get_value(self): return self.value
    def set_value(self, value): self.value = value
    def hasContent_(self):
        if (
            super(ProvenanceEntityRole, self).hasContent_()
        ):
            return True
        else:
            return False
    def export(self, outfile, level, namespace_='', name_='ProvenanceEntityRole', namespacedef_='', pretty_print=True):
        if pretty_print:
            eol_ = '\n'
        else:
            eol_ = ''
        if self.original_tagname_ is not None:
            name_ = self.original_tagname_
        showIndent(outfile, level, pretty_print)
        outfile.write('<%s%s%s' % (namespace_, name_, namespacedef_ and ' ' + namespacedef_ or '', ))
        already_processed = set()
        self.exportAttributes(outfile, level, already_processed, namespace_, name_='ProvenanceEntityRole')
        if self.hasContent_():
            outfile.write('>%s' % (eol_, ))
            self.exportChildren(outfile, level + 1, namespace_='', name_='ProvenanceEntityRole', pretty_print=pretty_print)
            showIndent(outfile, level, pretty_print)
            outfile.write('</%s%s>%s' % (namespace_, name_, eol_))
        else:
            outfile.write('/>%s' % (eol_, ))
    def exportAttributes(self, outfile, level, already_processed, namespace_='', name_='ProvenanceEntityRole'):
        super(ProvenanceEntityRole, self).exportAttributes(outfile, level, already_processed, namespace_, name_='ProvenanceEntityRole')
        if self.value is not None and 'value' not in already_processed:
            already_processed.add('value')
            outfile.write(' value=%s' % (quote_attrib(self.value), ))
    def exportChildren(self, outfile, level, namespace_='', name_='ProvenanceEntityRole', fromsubclass_=False, pretty_print=True):
        super(ProvenanceEntityRole, self).exportChildren(outfile, level, namespace_, name_, True, pretty_print=pretty_print)
    def to_etree(self, parent_element=None, name_='ProvenanceEntityRole', mapping_=None):
        element = super(ProvenanceEntityRole, self).to_etree(parent_element, name_, mapping_)
        if self.value is not None:
            element.set('value', self.value)
        if mapping_ is not None:
            mapping_[self] = element
        return element
    def build(self, node):
        already_processed = set()
        self.buildAttributes(node, node.attrib, already_processed)
        for child in node:
            nodeName_ = Tag_pattern_.match(child.tag).groups()[-1]
            self.buildChildren(child, node, nodeName_)
        return self
    def buildAttributes(self, node, attrs, already_processed):
        value = find_attr_value_('value', node)
        if value is not None and 'value' not in already_processed:
            already_processed.add('value')
            self.value = value
        super(ProvenanceEntityRole, self).buildAttributes(node, attrs, already_processed)
    def buildChildren(self, child_, node, nodeName_, fromsubclass_=False):
        super(ProvenanceEntityRole, self).buildChildren(child_, node, nodeName_, True)
        pass
# end class ProvenanceEntityRole



import server.fhir as supermod
from server.common import safe_attrgetter
from StringIO import StringIO
from operator import attrgetter
from .health_mixin import ExportXMLMixin

try:
    from flask import url_for
    RUN_FLASK=True
except:
    from .datastore import dumb_url_generate
    RUN_FLASK=False

class DiagnosticReport_Map:
    """
    The model mapping for the DiagnosticReport Resource
    """
    # Must be completed (i.e., have date)
    root_search=[('date_analysis', '!=', None)]

    # No other models, just use row id
    url_prefixes={}

    model_mapping={
            'gnuhealth.lab': {
                'subject': 'patient',
                'performer': 'pathologist',
                'conclusion': 'results', #or diagnosis?
                'codedDiagnosis': 'diagnosis',
                'result': 'critearea',
                'date': 'date_analysis',
                'test': 'test',
                'code': 'test.code',
                'name': 'test.name'}
                }
    resource_search_params={
                    '_id': 'token',
                    '_language': None,
                    'date': None,
                    'diagnosis': None, 
                    'identifier': None,
                    'image': None,
                    'issued': 'date',
                    'name': 'token',
                    'performer': 'reference',
                    'request': None,
                    'result': 'reference',
                    'service': None,
                    'specimen': None,
                    'status': None,
                    'subject': 'reference'}

    # Reference parameters to resource type
    chain_map={
            'subject': 'Patient',
            'result': 'Observation',
            'performer': 'Practitioner'}

    search_mapping={
                    '_id': ['id'],
                    'issued': ['date_analysis'],
                    'name': ['test.code'],
                    'name:text': ['test.name'],
                    'performer': ['pathologist'],
                    'result': ['critearea'],
                    'subject': ['patient']}

class health_DiagnosticReport(supermod.DiagnosticReport, DiagnosticReport_Map, ExportXMLMixin):
    """
    Class that manages the interface between FHIR Resource DiagnosticReport
    and GNU Health
    """
    def __init__(self, *args, **kwargs):
        rec = kwargs.pop('gnu_record', None)
        super(health_DiagnosticReport, self).__init__(*args, **kwargs)
        if rec:
            self.set_gnu_diagnostic_report(rec)

    def set_gnu_diagnostic_report(self, diagnostic_report):
        """Set the GNU Health model

        Keyword arguments:
        diagnostic_report -- the Health model
        """

        self.diagnostic_report = diagnostic_report
        self.model_type = self.diagnostic_report.__name__

        # Only certain models
        if self.model_type not in self.model_mapping:
            raise ValueError('Not a valid model')

        self.map = self.model_mapping[self.model_type]

        self.__import_from_gnu_diagnostic_report()
        self.__set_feed_info()

    def __import_from_gnu_diagnostic_report(self):
        """Import data from the model"""
        if self.diagnostic_report:
            self.set_conclusion(safe_attrgetter(self.diagnostic_report, self.map['conclusion']))
            self.set_subject(safe_attrgetter(self.diagnostic_report, self.map['subject']))
            self.set_performer(safe_attrgetter(self.diagnostic_report, self.map['performer']))
            self.set_issued(safe_attrgetter(self.diagnostic_report, self.map['date']))
            self.set_result(safe_attrgetter(self.diagnostic_report, self.map['result']))
            self.set_name(safe_attrgetter(self.diagnostic_report, self.map['test']))
            self.set_identifier(
                    safe_attrgetter(self.diagnostic_report, self.map['test']),
                    safe_attrgetter(self.diagnostic_report, 'patient.rec_name'),
                    safe_attrgetter(self.diagnostic_report, self.map['date']))

    def set_identifier(self, report, patient, date):
        """Extends superclass for convenience

        Keyword arguments:
        report -- the report (Health model)
        patient -- patient name
        date -- the date (datetime object)
        """

        if report and patient and date:
            label = '{0} for {1} on {2}'.format(report.name, patient, date.strftime('%Y/%m/%d'))
            if RUN_FLASK:
                value = url_for('dr_record', log_id=report.id)
            else:
                value = dumb_url_generate(['DiagnosticReport', report.id])
            ident = supermod.Identifier(
                        label=supermod.string(value=label),
                        value=supermod.string(value=value))
            super(health_DiagnosticReport, self).set_identifier(ident)

    def __set_feed_info(self):
        """Sets the feed-relevant data"""
        if self.diagnostic_report:
            if RUN_FLASK:
                uri = url_for('dr_record',
                                log_id=self.diagnostic_report.id,
                                _external=True)
            else:
                uri = dumb_url_generate(['DiagnosticReport',
                                self.diagnostic_report.id])
            self.feed={'id': uri,
                    'published': self.diagnostic_report.create_date,
                    'updated': self.diagnostic_report.write_date or self.diagnostic_report.create_date,
                    'title': self.diagnostic_report.rec_name
                        }

    def set_name(self, name):
        """Extends superclass for convenience

        Keyword arguments:
        name -- the test (Health model)
        """

        if name:
            conc = supermod.CodeableConcept()
            conc.coding=[supermod.Coding()]
            conc.coding[0].display=supermod.string(value=name.name)
            conc.coding[0].code = supermod.code(value=name.code)
            super(health_DiagnosticReport, self).set_name(conc)

        else:
            # If you don't know what is being tested,
            #   the data is useless
            raise ValueError('No test coding info')

    def set_issued(self, issued):
        """Extends superclass for convenience

        Keyword arguments:
        issued -- the date issued (datetime object)
        """

        if issued is not None:
            instant = supermod.instant(value=issued.strftime("%Y-%m-%dT%H:%M:%S"))
            super(health_DiagnosticReport, self).set_issued(instant)
        else:
            # If there is no date attached, this report is either not done
            #  or useless
            raise ValueError('No date')

    def set_result(self, result):
        """Extends superclass for convenience

        Keyword arguments:
        result -- the results
        """

        if result:
            for test in result:
                if RUN_FLASK:
                    uri = url_for('obs_record', log_id=test.id)
                else:
                    uri = dumb_url_generate(['Observation', test.id])
                display = test.rec_name
                ref=supermod.ResourceReference()
                ref.display = supermod.string(value=display)
                ref.reference = supermod.string(value=uri)
                super(health_DiagnosticReport, self).add_result(ref)

            # No data = useless
            if len(self.get_result()) == 0:
                raise ValueError('No data')

        else:
            # No data = useless
            raise ValueError('No data')

    def set_performer(self, performer):
        """Extends superclass for convenience

        Keyword arguments:
        performer -- the lab performer (Health model)
        """

        if performer:
            if RUN_FLASK:
                uri = url_for('hp_record', log_id=performer.id)
            else:
                uri = dumb_url_generate(['Practitioner', performer.id])
            display = performer.name.rec_name
            ref=supermod.ResourceReference()
            ref.display = supermod.string(value=display)
            ref.reference = supermod.string(value=uri)
            super(health_DiagnosticReport, self).set_performer(ref)

    def set_subject(self, subject):
        """Extends superclass for convenience

        Keyword arguments:
        subject -- the patient (Health model)
        """

        if subject:
            patient = attrgetter(self.map['subject'])(self.diagnostic_report)
            if RUN_FLASK:
                uri = url_for('pat_record', log_id=subject.id)
            else:
                uri = dumb_url_generate(['Patient', subject.id])
            display = subject.rec_name
            ref=supermod.ResourceReference()
            ref.display = supermod.string(value=display)
            ref.reference = supermod.string(value=uri)
            super(health_DiagnosticReport, self).set_subject(ref)

        else:
            # Without subject, useless information
            raise ValueError('No subject')

    def set_conclusion(self, conclusion):
        """Extends superclass for convenience

        Keyword arguments:
        conclusion -- the report's conclusion
        """

        if conclusion:
            c = supermod.string(value=conclusion)
            super(health_DiagnosticReport, self).set_conclusion(c)

supermod.DiagnosticReport.subclass=health_DiagnosticReport



"""

davcmd.py
---------

containts commands like copy, move, delete for normal
resources and collections

"""


import urllib.request, urllib.parse, urllib.error

from .utils import create_treelist, is_prefix
from .errors import *

def deltree(dc,uri,exclude={}):
    """ delete a tree of resources

    dc  -- dataclass to use
    uri -- root uri to delete
    exclude -- an optional list of uri:error_code pairs which should not
           be deleted.

    returns dict of uri:error_code tuples from which
    another method can create a multistatus xml element.

    Also note that we only know Depth=infinity thus we don't have
    to test for it.

    """

    tlist=create_treelist(dc,uri)
    result={}

    for i in range(len(tlist),0,-1):
        problem_uris=list(result.keys())
        element=tlist[i-1]

        # test here, if an element is a prefix of an uri which
        # generated an error before.
        # note that we walk here from childs to parents, thus
        # we cannot delete a parent if a child made a problem.
        # (see example in 8.6.2.1)
        ok=1
        for p in problem_uris:
            if is_prefix(element,p):
                ok=None
                break

            if not ok: continue

        # here we test for the exclude list which is the other way round!
        for p in list(exclude.keys()):
            if is_prefix(p,element):
                ok=None
                break

            if not ok: continue

        # now delete stuff
        try:
            delone(dc,element)
        except DAV_Error as error:
            (ec,dd) = error.args
            result[element]=ec

    return result

def delone(dc,uri):
    """ delete a single object """
    if dc.is_collection(uri):
        return dc.rmcol(uri)   # should be empty
    else:
        return dc.rm(uri)

###
### COPY
###

# helper function

def copy(dc,src,dst):
    """ only copy the element

    This is just a helper method factored out from copy and
    copytree. It will not handle the overwrite or depth header.

    """

    # destination should have been deleted before
    if dc.exists(dst): 
        raise DAV_Error(412)

    # source should exist also
    if not dc.exists(src): 
        raise DAV_NotFound

    if dc.is_collection(src):
        dc.copycol(src, dst) # an exception will be passed thru
    else:
        dc.copy(src, dst)  # an exception will be passed thru

# the main functions

def copyone(dc,src,dst,overwrite=None):
    """ copy one resource to a new destination """

    if overwrite and dc.exists(dst):
        delres = deltree(dc, dst)
    else:
        delres={}

    # if we cannot delete everything, then do not copy!
    if delres: 
        return delres

    try:
        copy(dc, src, dst)    # pass thru exceptions
    except DAV_Error as error:
        (ec, dd) = error.args
        return ec

def copytree(dc,src,dst,overwrite=None):
    """ copy a tree of resources to another location

    dc  -- dataclass to use
    src -- src uri from where to copy
    dst -- dst uri
    overwrite -- if 1 then delete dst uri before

    returns dict of uri:error_code tuples from which
    another method can create a multistatus xml element.

    """

    # first delete the destination resource
    if overwrite and dc.exists(dst):
        delres=deltree(dc,dst)
    else:
        delres={}

    # if we cannot delete everything, then do not copy!
    if delres: 
        return delres

    # get the tree we have to copy
    tlist = create_treelist(dc,src)
    result = {}

    # prepare destination URIs (get the prefix)
    dpath = urllib.parse.urlparse(dst)[2]

    for element in tlist:
        problem_uris = list(result.keys())

        # now URIs get longer and longer thus we have
        # to test if we had a parent URI which we were not
        # able to copy in problem_uris which is the prefix
        # of the actual element. If it is, then we cannot
        # copy this as well but do not generate another error.
        ok=1
        for p in problem_uris:
            if is_prefix(p,element):
                ok=None
                break

        if not ok: continue

        # now create the destination URI which corresponds to
        # the actual source URI. -> actual_dst
        # ("subtract" the base src from the URI and prepend the
        # dst prefix to it.)
        esrc=element.replace(src,b"")
        actual_dst=dpath+esrc

        # now copy stuff
        try:
            copy(dc,element,actual_dst)
        except DAV_Error as error:
            (ec,dd) = error.args
            result[element]=ec

    return result


###
### MOVE
###


def moveone(dc,src,dst,overwrite=None):
    """ move a single resource

    This is done by first copying it and then deleting
    the original.
    """

    # first copy it
    copyone(dc, src, dst, overwrite)

    # then delete it
    dc.rm(src)

def movetree(dc,src,dst,overwrite=None):
    """ move a collection

    This is done by first copying it and then deleting
    the original.

    PROBLEM: if something did not copy then we have a problem
    when deleting as the original might get deleted!
    """

    # first copy it
    res = copytree(dc,src,dst,overwrite)

    # then delete it
    res = deltree(dc,src,exclude=res)

    return res




from .base_classes import *
from .support_functions import *


class Medication(Resource):
    """Primarily used for identification and definition of Medication, but
    also covers ingredients and packaging.If the element is present,
    it must have either a @value, an @id, or extensions"""
    subclass = None
    superclass = Resource
    def __init__(self, id=None, extension=None, modifierExtension=None, language=None, text=None, contained=None, name=None, code=None, isBrand=None, manufacturer=None, kind=None, product=None, package=None):
        self.original_tagname_ = None
        super(Medication, self).__init__(id, extension, modifierExtension, language, text, contained, )
        self.name = name
        self.code = code
        self.isBrand = isBrand
        self.manufacturer = manufacturer
        self.kind = kind
        self.product = product
        self.package = package
    def factory(*args_, **kwargs_):
        if Medication.subclass:
            return Medication.subclass(*args_, **kwargs_)
        else:
            return Medication(*args_, **kwargs_)
    factory = staticmethod(factory)
    def get_name(self): return self.name
    def set_name(self, name): self.name = name
    def get_code(self): return self.code
    def set_code(self, code): self.code = code
    def get_isBrand(self): return self.isBrand
    def set_isBrand(self, isBrand): self.isBrand = isBrand
    def get_manufacturer(self): return self.manufacturer
    def set_manufacturer(self, manufacturer): self.manufacturer = manufacturer
    def get_kind(self): return self.kind
    def set_kind(self, kind): self.kind = kind
    def get_product(self): return self.product
    def set_product(self, product): self.product = product
    def get_package(self): return self.package
    def set_package(self, package): self.package = package
    def hasContent_(self):
        if (
            self.name is not None or
            self.code is not None or
            self.isBrand is not None or
            self.manufacturer is not None or
            self.kind is not None or
            self.product is not None or
            self.package is not None or
            super(Medication, self).hasContent_()
        ):
            return True
        else:
            return False
    def export(self, outfile, level, namespace_='', name_='Medication', namespacedef_='', pretty_print=True):
        if pretty_print:
            eol_ = '\n'
        else:
            eol_ = ''
        if self.original_tagname_ is not None:
            name_ = self.original_tagname_
        showIndent(outfile, level, pretty_print)
        outfile.write('<%s%s%s' % (namespace_, name_, namespacedef_ and ' ' + namespacedef_ or '', ))
        already_processed = set()
        self.exportAttributes(outfile, level, already_processed, namespace_, name_='Medication')
        if self.hasContent_():
            outfile.write('>%s' % (eol_, ))
            self.exportChildren(outfile, level + 1, namespace_='', name_='Medication', pretty_print=pretty_print)
            showIndent(outfile, level, pretty_print)
            outfile.write('</%s%s>%s' % (namespace_, name_, eol_))
        else:
            outfile.write('/>%s' % (eol_, ))
    def exportAttributes(self, outfile, level, already_processed, namespace_='', name_='Medication'):
        super(Medication, self).exportAttributes(outfile, level, already_processed, namespace_, name_='Medication')
    def exportChildren(self, outfile, level, namespace_='', name_='Medication', fromsubclass_=False, pretty_print=True):
        super(Medication, self).exportChildren(outfile, level, namespace_, name_, True, pretty_print=pretty_print)
        if pretty_print:
            eol_ = '\n'
        else:
            eol_ = ''
        if self.name is not None:
            self.name.export(outfile, level, namespace_, name_='name', pretty_print=pretty_print)
        if self.code is not None:
            self.code.export(outfile, level, namespace_, name_='code', pretty_print=pretty_print)
        if self.isBrand is not None:
            self.isBrand.export(outfile, level, namespace_, name_='isBrand', pretty_print=pretty_print)
        if self.manufacturer is not None:
            self.manufacturer.export(outfile, level, namespace_, name_='manufacturer', pretty_print=pretty_print)
        if self.kind is not None:
            self.kind.export(outfile, level, namespace_, name_='kind', pretty_print=pretty_print)
        if self.product is not None:
            self.product.export(outfile, level, namespace_, name_='product', pretty_print=pretty_print)
        if self.package is not None:
            self.package.export(outfile, level, namespace_, name_='package', pretty_print=pretty_print)
    def to_etree(self, parent_element=None, name_='Medication', mapping_=None):
        element = super(Medication, self).to_etree(parent_element, name_, mapping_)
        if self.name is not None:
            name_ = self.name
            name_.to_etree(element, name_='name', mapping_=mapping_)
        if self.code is not None:
            code_ = self.code
            code_.to_etree(element, name_='code', mapping_=mapping_)
        if self.isBrand is not None:
            isBrand_ = self.isBrand
            isBrand_.to_etree(element, name_='isBrand', mapping_=mapping_)
        if self.manufacturer is not None:
            manufacturer_ = self.manufacturer
            manufacturer_.to_etree(element, name_='manufacturer', mapping_=mapping_)
        if self.kind is not None:
            kind_ = self.kind
            kind_.to_etree(element, name_='kind', mapping_=mapping_)
        if self.product is not None:
            product_ = self.product
            product_.to_etree(element, name_='product', mapping_=mapping_)
        if self.package is not None:
            package_ = self.package
            package_.to_etree(element, name_='package', mapping_=mapping_)
        if mapping_ is not None:
            mapping_[self] = element
        return element
    def build(self, node):
        already_processed = set()
        self.buildAttributes(node, node.attrib, already_processed)
        for child in node:
            nodeName_ = Tag_pattern_.match(child.tag).groups()[-1]
            self.buildChildren(child, node, nodeName_)
        return self
    def buildAttributes(self, node, attrs, already_processed):
        super(Medication, self).buildAttributes(node, attrs, already_processed)
    def buildChildren(self, child_, node, nodeName_, fromsubclass_=False):
        if nodeName_ == 'name':
            obj_ = string.factory()
            obj_.build(child_)
            self.name = obj_
            obj_.original_tagname_ = 'name'
        elif nodeName_ == 'code':
            obj_ = CodeableConcept.factory()
            obj_.build(child_)
            self.code = obj_
            obj_.original_tagname_ = 'code'
        elif nodeName_ == 'isBrand':
            obj_ = boolean.factory()
            obj_.build(child_)
            self.isBrand = obj_
            obj_.original_tagname_ = 'isBrand'
        elif nodeName_ == 'manufacturer':
            obj_ = ResourceReference.factory()
            obj_.build(child_)
            self.manufacturer = obj_
            obj_.original_tagname_ = 'manufacturer'
        elif nodeName_ == 'kind':
            obj_ = MedicationKind.factory()
            obj_.build(child_)
            self.kind = obj_
            obj_.original_tagname_ = 'kind'
        elif nodeName_ == 'product':
            obj_ = Medication_Product.factory()
            obj_.build(child_)
            self.product = obj_
            obj_.original_tagname_ = 'product'
        elif nodeName_ == 'package':
            obj_ = Medication_Package.factory()
            obj_.build(child_)
            self.package = obj_
            obj_.original_tagname_ = 'package'
        super(Medication, self).buildChildren(child_, node, nodeName_, True)
# end class Medication


class Medication_Product(BackboneElement):
    """Primarily used for identification and definition of Medication, but
    also covers ingredients and packaging."""
    subclass = None
    superclass = BackboneElement
    def __init__(self, id=None, extension=None, modifierExtension=None, form=None, ingredient=None):
        self.original_tagname_ = None
        super(Medication_Product, self).__init__(id, extension, modifierExtension, )
        self.form = form
        if ingredient is None:
            self.ingredient = []
        else:
            self.ingredient = ingredient
    def factory(*args_, **kwargs_):
        if Medication_Product.subclass:
            return Medication_Product.subclass(*args_, **kwargs_)
        else:
            return Medication_Product(*args_, **kwargs_)
    factory = staticmethod(factory)
    def get_form(self): return self.form
    def set_form(self, form): self.form = form
    def get_ingredient(self): return self.ingredient
    def set_ingredient(self, ingredient): self.ingredient = ingredient
    def add_ingredient(self, value): self.ingredient.append(value)
    def insert_ingredient(self, index, value): self.ingredient[index] = value
    def hasContent_(self):
        if (
            self.form is not None or
            self.ingredient or
            super(Medication_Product, self).hasContent_()
        ):
            return True
        else:
            return False
    def export(self, outfile, level, namespace_='', name_='Medication.Product', namespacedef_='', pretty_print=True):
        if pretty_print:
            eol_ = '\n'
        else:
            eol_ = ''
        if self.original_tagname_ is not None:
            name_ = self.original_tagname_
        showIndent(outfile, level, pretty_print)
        outfile.write('<%s%s%s' % (namespace_, name_, namespacedef_ and ' ' + namespacedef_ or '', ))
        already_processed = set()
        self.exportAttributes(outfile, level, already_processed, namespace_, name_='Medication.Product')
        if self.hasContent_():
            outfile.write('>%s' % (eol_, ))
            self.exportChildren(outfile, level + 1, namespace_='', name_='Medication.Product', pretty_print=pretty_print)
            showIndent(outfile, level, pretty_print)
            outfile.write('</%s%s>%s' % (namespace_, name_, eol_))
        else:
            outfile.write('/>%s' % (eol_, ))
    def exportAttributes(self, outfile, level, already_processed, namespace_='', name_='Medication.Product'):
        super(Medication_Product, self).exportAttributes(outfile, level, already_processed, namespace_, name_='Medication.Product')
    def exportChildren(self, outfile, level, namespace_='', name_='Medication.Product', fromsubclass_=False, pretty_print=True):
        super(Medication_Product, self).exportChildren(outfile, level, namespace_, name_, True, pretty_print=pretty_print)
        if pretty_print:
            eol_ = '\n'
        else:
            eol_ = ''
        if self.form is not None:
            self.form.export(outfile, level, namespace_, name_='form', pretty_print=pretty_print)
        for ingredient_ in self.ingredient:
            ingredient_.export(outfile, level, namespace_, name_='ingredient', pretty_print=pretty_print)
    def to_etree(self, parent_element=None, name_='Medication.Product', mapping_=None):
        element = super(Medication_Product, self).to_etree(parent_element, name_, mapping_)
        if self.form is not None:
            form_ = self.form
            form_.to_etree(element, name_='form', mapping_=mapping_)
        for ingredient_ in self.ingredient:
            ingredient_.to_etree(element, name_='ingredient', mapping_=mapping_)
        if mapping_ is not None:
            mapping_[self] = element
        return element
    def build(self, node):
        already_processed = set()
        self.buildAttributes(node, node.attrib, already_processed)
        for child in node:
            nodeName_ = Tag_pattern_.match(child.tag).groups()[-1]
            self.buildChildren(child, node, nodeName_)
        return self
    def buildAttributes(self, node, attrs, already_processed):
        super(Medication_Product, self).buildAttributes(node, attrs, already_processed)
    def buildChildren(self, child_, node, nodeName_, fromsubclass_=False):
        if nodeName_ == 'form':
            obj_ = CodeableConcept.factory()
            obj_.build(child_)
            self.form = obj_
            obj_.original_tagname_ = 'form'
        elif nodeName_ == 'ingredient':
            obj_ = Medication_Ingredient.factory()
            obj_.build(child_)
            self.ingredient.append(obj_)
            obj_.original_tagname_ = 'ingredient'
        super(Medication_Product, self).buildChildren(child_, node, nodeName_, True)
# end class Medication_Product


class Medication_Ingredient(BackboneElement):
    """Primarily used for identification and definition of Medication, but
    also covers ingredients and packaging."""
    subclass = None
    superclass = BackboneElement
    def __init__(self, id=None, extension=None, modifierExtension=None, item=None, amount=None):
        self.original_tagname_ = None
        super(Medication_Ingredient, self).__init__(id, extension, modifierExtension, )
        self.item = item
        self.amount = amount
    def factory(*args_, **kwargs_):
        if Medication_Ingredient.subclass:
            return Medication_Ingredient.subclass(*args_, **kwargs_)
        else:
            return Medication_Ingredient(*args_, **kwargs_)
    factory = staticmethod(factory)
    def get_item(self): return self.item
    def set_item(self, item): self.item = item
    def get_amount(self): return self.amount
    def set_amount(self, amount): self.amount = amount
    def hasContent_(self):
        if (
            self.item is not None or
            self.amount is not None or
            super(Medication_Ingredient, self).hasContent_()
        ):
            return True
        else:
            return False
    def export(self, outfile, level, namespace_='', name_='Medication.Ingredient', namespacedef_='', pretty_print=True):
        if pretty_print:
            eol_ = '\n'
        else:
            eol_ = ''
        if self.original_tagname_ is not None:
            name_ = self.original_tagname_
        showIndent(outfile, level, pretty_print)
        outfile.write('<%s%s%s' % (namespace_, name_, namespacedef_ and ' ' + namespacedef_ or '', ))
        already_processed = set()
        self.exportAttributes(outfile, level, already_processed, namespace_, name_='Medication.Ingredient')
        if self.hasContent_():
            outfile.write('>%s' % (eol_, ))
            self.exportChildren(outfile, level + 1, namespace_='', name_='Medication.Ingredient', pretty_print=pretty_print)
            showIndent(outfile, level, pretty_print)
            outfile.write('</%s%s>%s' % (namespace_, name_, eol_))
        else:
            outfile.write('/>%s' % (eol_, ))
    def exportAttributes(self, outfile, level, already_processed, namespace_='', name_='Medication.Ingredient'):
        super(Medication_Ingredient, self).exportAttributes(outfile, level, already_processed, namespace_, name_='Medication.Ingredient')
    def exportChildren(self, outfile, level, namespace_='', name_='Medication.Ingredient', fromsubclass_=False, pretty_print=True):
        super(Medication_Ingredient, self).exportChildren(outfile, level, namespace_, name_, True, pretty_print=pretty_print)
        if pretty_print:
            eol_ = '\n'
        else:
            eol_ = ''
        if self.item is not None:
            self.item.export(outfile, level, namespace_, name_='item', pretty_print=pretty_print)
        if self.amount is not None:
            self.amount.export(outfile, level, namespace_, name_='amount', pretty_print=pretty_print)
    def to_etree(self, parent_element=None, name_='Medication.Ingredient', mapping_=None):
        element = super(Medication_Ingredient, self).to_etree(parent_element, name_, mapping_)
        if self.item is not None:
            item_ = self.item
            item_.to_etree(element, name_='item', mapping_=mapping_)
        if self.amount is not None:
            amount_ = self.amount
            amount_.to_etree(element, name_='amount', mapping_=mapping_)
        if mapping_ is not None:
            mapping_[self] = element
        return element
    def build(self, node):
        already_processed = set()
        self.buildAttributes(node, node.attrib, already_processed)
        for child in node:
            nodeName_ = Tag_pattern_.match(child.tag).groups()[-1]
            self.buildChildren(child, node, nodeName_)
        return self
    def buildAttributes(self, node, attrs, already_processed):
        super(Medication_Ingredient, self).buildAttributes(node, attrs, already_processed)
    def buildChildren(self, child_, node, nodeName_, fromsubclass_=False):
        if nodeName_ == 'item':
            obj_ = ResourceReference.factory()
            obj_.build(child_)
            self.item = obj_
            obj_.original_tagname_ = 'item'
        elif nodeName_ == 'amount':
            obj_ = Ratio.factory()
            obj_.build(child_)
            self.amount = obj_
            obj_.original_tagname_ = 'amount'
        super(Medication_Ingredient, self).buildChildren(child_, node, nodeName_, True)
# end class Medication_Ingredient


class Medication_Package(BackboneElement):
    """Primarily used for identification and definition of Medication, but
    also covers ingredients and packaging."""
    subclass = None
    superclass = BackboneElement
    def __init__(self, id=None, extension=None, modifierExtension=None, container=None, content=None):
        self.original_tagname_ = None
        super(Medication_Package, self).__init__(id, extension, modifierExtension, )
        self.container = container
        if content is None:
            self.content = []
        else:
            self.content = content
    def factory(*args_, **kwargs_):
        if Medication_Package.subclass:
            return Medication_Package.subclass(*args_, **kwargs_)
        else:
            return Medication_Package(*args_, **kwargs_)
    factory = staticmethod(factory)
    def get_container(self): return self.container
    def set_container(self, container): self.container = container
    def get_content(self): return self.content
    def set_content(self, content): self.content = content
    def add_content(self, value): self.content.append(value)
    def insert_content(self, index, value): self.content[index] = value
    def hasContent_(self):
        if (
            self.container is not None or
            self.content or
            super(Medication_Package, self).hasContent_()
        ):
            return True
        else:
            return False
    def export(self, outfile, level, namespace_='', name_='Medication.Package', namespacedef_='', pretty_print=True):
        if pretty_print:
            eol_ = '\n'
        else:
            eol_ = ''
        if self.original_tagname_ is not None:
            name_ = self.original_tagname_
        showIndent(outfile, level, pretty_print)
        outfile.write('<%s%s%s' % (namespace_, name_, namespacedef_ and ' ' + namespacedef_ or '', ))
        already_processed = set()
        self.exportAttributes(outfile, level, already_processed, namespace_, name_='Medication.Package')
        if self.hasContent_():
            outfile.write('>%s' % (eol_, ))
            self.exportChildren(outfile, level + 1, namespace_='', name_='Medication.Package', pretty_print=pretty_print)
            showIndent(outfile, level, pretty_print)
            outfile.write('</%s%s>%s' % (namespace_, name_, eol_))
        else:
            outfile.write('/>%s' % (eol_, ))
    def exportAttributes(self, outfile, level, already_processed, namespace_='', name_='Medication.Package'):
        super(Medication_Package, self).exportAttributes(outfile, level, already_processed, namespace_, name_='Medication.Package')
    def exportChildren(self, outfile, level, namespace_='', name_='Medication.Package', fromsubclass_=False, pretty_print=True):
        super(Medication_Package, self).exportChildren(outfile, level, namespace_, name_, True, pretty_print=pretty_print)
        if pretty_print:
            eol_ = '\n'
        else:
            eol_ = ''
        if self.container is not None:
            self.container.export(outfile, level, namespace_, name_='container', pretty_print=pretty_print)
        for content_ in self.content:
            content_.export(outfile, level, namespace_, name_='content', pretty_print=pretty_print)
    def to_etree(self, parent_element=None, name_='Medication.Package', mapping_=None):
        element = super(Medication_Package, self).to_etree(parent_element, name_, mapping_)
        if self.container is not None:
            container_ = self.container
            container_.to_etree(element, name_='container', mapping_=mapping_)
        for content_ in self.content:
            content_.to_etree(element, name_='content', mapping_=mapping_)
        if mapping_ is not None:
            mapping_[self] = element
        return element
    def build(self, node):
        already_processed = set()
        self.buildAttributes(node, node.attrib, already_processed)
        for child in node:
            nodeName_ = Tag_pattern_.match(child.tag).groups()[-1]
            self.buildChildren(child, node, nodeName_)
        return self
    def buildAttributes(self, node, attrs, already_processed):
        super(Medication_Package, self).buildAttributes(node, attrs, already_processed)
    def buildChildren(self, child_, node, nodeName_, fromsubclass_=False):
        if nodeName_ == 'container':
            obj_ = CodeableConcept.factory()
            obj_.build(child_)
            self.container = obj_
            obj_.original_tagname_ = 'container'
        elif nodeName_ == 'content':
            obj_ = Medication_Content.factory()
            obj_.build(child_)
            self.content.append(obj_)
            obj_.original_tagname_ = 'content'
        super(Medication_Package, self).buildChildren(child_, node, nodeName_, True)
# end class Medication_Package


class Medication_Content(BackboneElement):
    """Primarily used for identification and definition of Medication, but
    also covers ingredients and packaging."""
    subclass = None
    superclass = BackboneElement
    def __init__(self, id=None, extension=None, modifierExtension=None, item=None, amount=None):
        self.original_tagname_ = None
        super(Medication_Content, self).__init__(id, extension, modifierExtension, )
        self.item = item
        self.amount = amount
    def factory(*args_, **kwargs_):
        if Medication_Content.subclass:
            return Medication_Content.subclass(*args_, **kwargs_)
        else:
            return Medication_Content(*args_, **kwargs_)
    factory = staticmethod(factory)
    def get_item(self): return self.item
    def set_item(self, item): self.item = item
    def get_amount(self): return self.amount
    def set_amount(self, amount): self.amount = amount
    def hasContent_(self):
        if (
            self.item is not None or
            self.amount is not None or
            super(Medication_Content, self).hasContent_()
        ):
            return True
        else:
            return False
    def export(self, outfile, level, namespace_='', name_='Medication.Content', namespacedef_='', pretty_print=True):
        if pretty_print:
            eol_ = '\n'
        else:
            eol_ = ''
        if self.original_tagname_ is not None:
            name_ = self.original_tagname_
        showIndent(outfile, level, pretty_print)
        outfile.write('<%s%s%s' % (namespace_, name_, namespacedef_ and ' ' + namespacedef_ or '', ))
        already_processed = set()
        self.exportAttributes(outfile, level, already_processed, namespace_, name_='Medication.Content')
        if self.hasContent_():
            outfile.write('>%s' % (eol_, ))
            self.exportChildren(outfile, level + 1, namespace_='', name_='Medication.Content', pretty_print=pretty_print)
            showIndent(outfile, level, pretty_print)
            outfile.write('</%s%s>%s' % (namespace_, name_, eol_))
        else:
            outfile.write('/>%s' % (eol_, ))
    def exportAttributes(self, outfile, level, already_processed, namespace_='', name_='Medication.Content'):
        super(Medication_Content, self).exportAttributes(outfile, level, already_processed, namespace_, name_='Medication.Content')
    def exportChildren(self, outfile, level, namespace_='', name_='Medication.Content', fromsubclass_=False, pretty_print=True):
        super(Medication_Content, self).exportChildren(outfile, level, namespace_, name_, True, pretty_print=pretty_print)
        if pretty_print:
            eol_ = '\n'
        else:
            eol_ = ''
        if self.item is not None:
            self.item.export(outfile, level, namespace_, name_='item', pretty_print=pretty_print)
        if self.amount is not None:
            self.amount.export(outfile, level, namespace_, name_='amount', pretty_print=pretty_print)
    def to_etree(self, parent_element=None, name_='Medication.Content', mapping_=None):
        element = super(Medication_Content, self).to_etree(parent_element, name_, mapping_)
        if self.item is not None:
            item_ = self.item
            item_.to_etree(element, name_='item', mapping_=mapping_)
        if self.amount is not None:
            amount_ = self.amount
            amount_.to_etree(element, name_='amount', mapping_=mapping_)
        if mapping_ is not None:
            mapping_[self] = element
        return element
    def build(self, node):
        already_processed = set()
        self.buildAttributes(node, node.attrib, already_processed)
        for child in node:
            nodeName_ = Tag_pattern_.match(child.tag).groups()[-1]
            self.buildChildren(child, node, nodeName_)
        return self
    def buildAttributes(self, node, attrs, already_processed):
        super(Medication_Content, self).buildAttributes(node, attrs, already_processed)
    def buildChildren(self, child_, node, nodeName_, fromsubclass_=False):
        if nodeName_ == 'item':
            obj_ = ResourceReference.factory()
            obj_.build(child_)
            self.item = obj_
            obj_.original_tagname_ = 'item'
        elif nodeName_ == 'amount':
            obj_ = Quantity.factory()
            obj_.build(child_)
            self.amount = obj_
            obj_.original_tagname_ = 'amount'
        super(Medication_Content, self).buildChildren(child_, node, nodeName_, True)
# end class Medication_Content


class MedicationKind(Element):
    """Whether the medication is a product or a packageIf the element is
    present, it must have either a @value, an @id, or extensions"""
    subclass = None
    superclass = Element
    def __init__(self, id=None, extension=None, value=None):
        self.original_tagname_ = None
        super(MedicationKind, self).__init__(id, extension, )
        self.value = _cast(None, value)
    def factory(*args_, **kwargs_):
        if MedicationKind.subclass:
            return MedicationKind.subclass(*args_, **kwargs_)
        else:
            return MedicationKind(*args_, **kwargs_)
    factory = staticmethod(factory)
    def get_value(self): return self.value
    def set_value(self, value): self.value = value
    def hasContent_(self):
        if (
            super(MedicationKind, self).hasContent_()
        ):
            return True
        else:
            return False
    def export(self, outfile, level, namespace_='', name_='MedicationKind', namespacedef_='', pretty_print=True):
        if pretty_print:
            eol_ = '\n'
        else:
            eol_ = ''
        if self.original_tagname_ is not None:
            name_ = self.original_tagname_
        showIndent(outfile, level, pretty_print)
        outfile.write('<%s%s%s' % (namespace_, name_, namespacedef_ and ' ' + namespacedef_ or '', ))
        already_processed = set()
        self.exportAttributes(outfile, level, already_processed, namespace_, name_='MedicationKind')
        if self.hasContent_():
            outfile.write('>%s' % (eol_, ))
            self.exportChildren(outfile, level + 1, namespace_='', name_='MedicationKind', pretty_print=pretty_print)
            showIndent(outfile, level, pretty_print)
            outfile.write('</%s%s>%s' % (namespace_, name_, eol_))
        else:
            outfile.write('/>%s' % (eol_, ))
    def exportAttributes(self, outfile, level, already_processed, namespace_='', name_='MedicationKind'):
        super(MedicationKind, self).exportAttributes(outfile, level, already_processed, namespace_, name_='MedicationKind')
        if self.value is not None and 'value' not in already_processed:
            already_processed.add('value')
            outfile.write(' value=%s' % (quote_attrib(self.value), ))
    def exportChildren(self, outfile, level, namespace_='', name_='MedicationKind', fromsubclass_=False, pretty_print=True):
        super(MedicationKind, self).exportChildren(outfile, level, namespace_, name_, True, pretty_print=pretty_print)
    def to_etree(self, parent_element=None, name_='MedicationKind', mapping_=None):
        element = super(MedicationKind, self).to_etree(parent_element, name_, mapping_)
        if self.value is not None:
            element.set('value', self.value)
        if mapping_ is not None:
            mapping_[self] = element
        return element
    def build(self, node):
        already_processed = set()
        self.buildAttributes(node, node.attrib, already_processed)
        for child in node:
            nodeName_ = Tag_pattern_.match(child.tag).groups()[-1]
            self.buildChildren(child, node, nodeName_)
        return self
    def buildAttributes(self, node, attrs, already_processed):
        value = find_attr_value_('value', node)
        if value is not None and 'value' not in already_processed:
            already_processed.add('value')
            self.value = value
        super(MedicationKind, self).buildAttributes(node, attrs, already_processed)
    def buildChildren(self, child_, node, nodeName_, fromsubclass_=False):
        super(MedicationKind, self).buildChildren(child_, node, nodeName_, True)
        pass
# end class MedicationKind



from trytond.pool import Pool
from .health_ophthalmology import *

def register():
    Pool.register(
        OphthalmologyEvaluation,
        OphthalmologyFindings,    
        module='health_ophthalmology', type_='model')



from math import ceil
from werkzeug.urls import Href, url_quote
from datetime import datetime
from StringIO import StringIO
import server.fhir as supermod

class Bundle(supermod.FeedType):
    """A bundle is an atom feed

    Subclasses the FeedType class but extends many
    functions for convenience and ease of use. In additon,
    adds some new functions and properties.
    """

    def __init__(self, *args, **kwargs):
        """Extends __init__ to set sane default values
        and add new attributes

        Keyword arguments:
        request -- Request object (required)
        total -- total results (required)
        title -- feed title (optional, default = Search results)
        author --  feed author (optional, default = GNU Health)
        updated -- feed updated (optional, default = utcnow())
        """

        self.request=kwargs.pop('request', None)
        id_=kwargs.pop('id', getattr(self.request, 'url'))
        page = int(self.request.args.get('page', 1)) or 1
        per_page= int(self.request.args.get('_count', 10)) or 10
        total = kwargs.pop('total', None)
        if total is None: raise ValueError('Total results required')
        title = kwargs.pop('title', 'Search results')
        author = kwargs.pop('author', 'GNU Health')
        updated = kwargs.pop('updated', datetime.utcnow())
        super(Bundle, self).__init__(*args, **kwargs)

        self.set_id(id_)
        self.set_title(title)
        self.set_author(author)
        self.set_updated(updated)
        self.set_totalResults(total)
        self.set_link(page, per_page, total)

    @property
    def entries(self):
        """Current entries"""
        return self.entry

    def __len__(self):
        return len(self.entry)

    def __getitem__(self, index):
        if index >= len(self):
            raise IndexError
        return self.entry[index]

    def set_id(self, id_):
        """Extends set_id function to properly escape ids

        Keyword arguments:
        id_ -- the feed id
        """

        if id_:
            i = supermod.IdType()
            i.set_valueOf_(url_quote(id_))
            super(Bundle, self).set_id([i])

    def set_author(self, author):
        """Extends set_author for string

        Keyword arguments:
        author -- the feed author
        """

        if author:
            a = supermod.PersonType()
            a.add_name(str(author))
            super(Bundle, self).set_author([a])

    def set_title(self, title):
        """Extends set_title for string

        Keyword arguments:
        title -- the feed title
        """

        if title:
            t = supermod.TextType(valueOf_=str(title))
            t.content_.append(t.mixedclass_(
                    supermod.MixedContainer.CategoryText,
                    supermod.MixedContainer.TypeNone, '', str(title)))
            super(Bundle, self).set_title([t])

    def set_updated(self, updated):
        """Extends set_updated for convenience

        Keyword arguments:
        updated -- feed updated datetime object
        """

        if updated is not None:
            u = supermod.DateTimeType()
            u.set_valueOf_(updated)
            super(Bundle, self).set_updated([u])

    def set_totalResults(self, total):
        """Extends set_totalResults for str/int

        Keyword argument:
        total -- total results
        """

        if total is not None:
            super(Bundle, self).set_totalResults([int(total)])

    def set_link(self, page, per_page, total):
        """Extends set_link for convenience, adds
        rel, prev, self, next, and last when applicable

        Keyword arguments:
        page -- current page
        per_page -- items per page
        total -- total results
        """

        links = []
        total_pages= int(ceil(float(total)/per_page))
        args = self.request.args.copy()
        href = Href(self.request.base_url)

        # self link
        l = supermod.linkType()
        l.rel = 'self'
        f = {k:v for k,v in args.items() if k not in ['_count', 'page']}
        l.href = href(f)
        links.append(l)

        if total_pages > 1:

            # first link
            if page > 1:
                args['page'] = 1
                l = supermod.linkType()
                l.rel = 'first'
                l.href = href(args)
                links.append(l)

            # last link
            if page < total_pages:
                args['page'] = total_pages
                l = supermod.linkType()
                l.rel = 'last'
                l.href = href(args)
                links.append(l)

            # next link
            if page < total_pages - 1:      # last link already generated
                args['page'] = page + 1
                l = supermod.linkType()
                l.rel = 'next'
                l.href = href(args)
                links.append(l)

            # prev link
            if page > 2:
                args['page'] = page-1
                l = supermod.linkType()
                l.rel = 'previous'
                l.href = href(args)
                links.append(l)

        super(Bundle, self).set_link(links)

    def add_entry(self, entry):
        """Extends add_entry for FHIR class instance

        Keyword arguments:
        entry -- entry to add (FHIR class object)
        """

        if entry.feed:
            e = supermod.EntryType()
            t=supermod.TextType(valueOf_=entry.feed['title'])
            t.content_.append(t.mixedclass_(
                    supermod.MixedContainer.CategoryText,
                    supermod.MixedContainer.TypeNone, '', entry.feed['title']))
            e.title=[t]
            e.id = [supermod.IdType(valueOf_=entry.feed['id'])]
            e.updated = [supermod.DateTimeType(valueOf_=entry.feed['updated'])]
            e.published = [supermod.DateTimeType(valueOf_=entry.feed['published'])]
            ct = supermod.ContentType(type_='text/xml') # only xml for now

            if isinstance(entry, supermod.Binary):
                t = 'Binary'
                ct.set_Binary(entry)
            elif isinstance(entry, supermod.AdverseReaction):
                t = 'AdverseReaction'
                ct.set_AdverseReaction(entry)
            elif isinstance(entry, supermod.Alert):
                t = 'Alert'
                ct.set_Alert(entry)
            elif isinstance(entry, supermod.AllergyIntolerance):
                t = 'AllergyIntolerance'
                ct.set_AllergyIntolerance(entry)
            elif isinstance(entry, supermod.CarePlan):
                t = 'CarePlan'
                ct.set_CarePlan(entry)
            elif isinstance(entry, supermod.Composition):
                t = 'Composition'
                ct.set_Composition(entry)
            elif isinstance(entry, supermod.ConceptMap):
                t = 'ConceptMap'
                ct.set_ConceptMap(entry)
            elif isinstance(entry, supermod.Condition):
                t = 'Condition'
                ct.set_Condition(entry)
            elif isinstance(entry, supermod.Conformance):
                t = 'Conformance'
                ct.set_Conformance(entry)
            elif isinstance(entry, supermod.Device):
                t = 'Device'
                ct.set_Device(entry)
            elif isinstance(entry, supermod.DeviceObservationReport):
                t = 'DeviceObservationReport'
                ct.set_DeviceObservationReport(entry)
            elif isinstance(entry, supermod.DiagnosticOrder):
                t = 'DiagnosticOrder'
                ct.set_DiagnosticOrder(entry)
            elif isinstance(entry, supermod.DiagnosticReport):
                t = 'DiagnosticReport'
                ct.set_DiagnosticReport(entry)
            elif isinstance(entry, supermod.DocumentManifest):
                t = 'DocumentManifest'
                ct.set_DocumentManifest(entry)
            elif isinstance(entry, supermod.DocumentReference):
                t = 'DocumentReference'
                ct.set_DocumentReference(entry)
            elif isinstance(entry, supermod.Encounter):
                t = 'Encounter'
                ct.set_Encounter(entry)
            elif isinstance(entry, supermod.FamilyHistory):
                t = 'FamilyHistory'
                ct.set_FamilyHistory(entry)
            elif isinstance(entry, supermod.Group):
                t = 'Group'
                ct.set_Group(entry)
            elif isinstance(entry, supermod.ImagingStudy):
                t = 'ImagingStudy'
                ct.set_ImagingStudy(entry)
            elif isinstance(entry, supermod.Immunization):
                t = 'Immunization'
                ct.set_Immunization(entry)
            elif isinstance(entry, supermod.ImmunizationRecommendation):
                t = 'ImmunizationRecommendation'
                ct.set_ImmunizationRecommendation(entry)
            elif isinstance(entry, supermod.List):
                t = 'List'
                ct.set_List(entry)
            elif isinstance(entry, supermod.Location):
                t = 'Location'
                ct.set_Location(entry)
            elif isinstance(entry, supermod.Media):
                t = 'Media'
                ct.set_Media(entry)
            elif isinstance(entry, supermod.Medication):
                t = 'Medication'
                ct.set_Medication(entry)
            elif isinstance(entry, supermod.MedicationAdministration):
                t = 'MedicationAdministration'
                ct.set_MedicationAdministration(entry)
            elif isinstance(entry, supermod.MedicationDispense):
                t = 'MedicationDispense'
                ct.set_MedicationDispense(entry)
            elif isinstance(entry, supermod.MedicationPrescription):
                t = 'MedicationPrescription'
                ct.set_MedicationPrescription(entry)
            elif isinstance(entry, supermod.MedicationStatement):
                t = 'MedicationStatement'
                ct.set_MedicationStatement(entry)
            elif isinstance(entry, supermod.MessageHeader):
                t = 'MessageHeader'
                ct.set_MessageHeader(entry)
            elif isinstance(entry, supermod.Observation):
                t = 'Observation'
                ct.set_Observation(entry)
            elif isinstance(entry, supermod.OperationOutcome):
                t = 'OperationOutcome'
                ct.set_OperationOutcome(entry)
            elif isinstance(entry, supermod.Order):
                t = 'Order'
                ct.set_Order(entry)
            elif isinstance(entry, supermod.OrderResponse):
                t = 'OrderResponse'
                ct.set_OrderResponse(entry)
            elif isinstance(entry, supermod.Organization):
                t = 'Organization'
                ct.set_Organization(entry)
            elif isinstance(entry, supermod.Other):
                t = 'Other'
                ct.set_Other(entry)
            elif isinstance(entry, supermod.Patient):
                t = 'Patient'
                ct.set_Patient(entry)
            elif isinstance(entry, supermod.Practitioner):
                t = 'Practitioner'
                ct.set_Practitioner(entry)
            elif isinstance(entry, supermod.Procedure):
                t = 'Procedure'
                ct.set_Procedure(entry)
            elif isinstance(entry, supermod.Profile):
                t = 'Profile'
                ct.set_Profile(entry)
            elif isinstance(entry, supermod.Provenance):
                t = 'Provenance'
                ct.set_Provenance(entry)
            elif isinstance(entry, supermod.Query):
                t = 'Query'
                ct.set_Query(entry)
            elif isinstance(entry, supermod.Questionnaire):
                t = 'Questionnaire'
                ct.set_Questionnaire(entry)
            elif isinstance(entry, supermod.RelatedPerson):
                t = 'RelatedPerson'
                ct.set_RelatedPerson(entry)
            elif isinstance(entry, supermod.SecurityEvent):
                t = 'SecurityEvent'
                ct.set_SecurityEvent(entry)
            elif isinstance(entry, supermod.Specimen):
                t = 'Specimen'
                ct.set_Specimen(entry)
            elif isinstance(entry, supermod.Substance):
                t = 'Substance'
                ct.set_Supply(entry)
            elif isinstance(entry, supermod.ValueSet):
                t = 'ValueSet'
                ct.set_ValueSet(entry)
            else:
                raise TypeError('Unknown FHIR class')

            # Now tell the class what type it is holding
            ct.content_.append(
                    ct.mixedclass_(
                        supermod.MixedContainer.CategoryComplex,
                        supermod.MixedContainer.TypeNone, t, entry))

            e.content = [ct]
            super(Bundle, self).add_entry(e)

    def export_to_xml_string(self):
        """Export to atom format"""
        i = StringIO()
        self.export(i,
                    level=0,
                    namespace_='',
                    name_='feed',
                    namespacedef_='xmlns="http://www.w3.org/2005/Atom" xmlns:os="http://a9.com/-/spec/opensearch/1.1/"')
        return i.getvalue()

supermod.FeedType.subclass = Bundle



from .base_classes import *
from .support_functions import *


class OperationOutcome(Resource):
    """A collection of error, warning or information messages that result
    from a system action.If the element is present, it must have
    either a @value, an @id, or extensions"""
    subclass = None
    superclass = Resource
    def __init__(self, id=None, extension=None, modifierExtension=None, language=None, text=None, contained=None, issue=None):
        self.original_tagname_ = None
        super(OperationOutcome, self).__init__(id, extension, modifierExtension, language, text, contained, )
        if issue is None:
            self.issue = []
        else:
            self.issue = issue
    def factory(*args_, **kwargs_):
        if OperationOutcome.subclass:
            return OperationOutcome.subclass(*args_, **kwargs_)
        else:
            return OperationOutcome(*args_, **kwargs_)
    factory = staticmethod(factory)
    def get_issue(self): return self.issue
    def set_issue(self, issue): self.issue = issue
    def add_issue(self, value): self.issue.append(value)
    def insert_issue(self, index, value): self.issue[index] = value
    def hasContent_(self):
        if (
            self.issue or
            super(OperationOutcome, self).hasContent_()
        ):
            return True
        else:
            return False
    def export(self, outfile, level, namespace_='', name_='OperationOutcome', namespacedef_='', pretty_print=True):
        if pretty_print:
            eol_ = '\n'
        else:
            eol_ = ''
        if self.original_tagname_ is not None:
            name_ = self.original_tagname_
        showIndent(outfile, level, pretty_print)
        outfile.write('<%s%s%s' % (namespace_, name_, namespacedef_ and ' ' + namespacedef_ or '', ))
        already_processed = set()
        self.exportAttributes(outfile, level, already_processed, namespace_, name_='OperationOutcome')
        if self.hasContent_():
            outfile.write('>%s' % (eol_, ))
            self.exportChildren(outfile, level + 1, namespace_='', name_='OperationOutcome', pretty_print=pretty_print)
            showIndent(outfile, level, pretty_print)
            outfile.write('</%s%s>%s' % (namespace_, name_, eol_))
        else:
            outfile.write('/>%s' % (eol_, ))
    def exportAttributes(self, outfile, level, already_processed, namespace_='', name_='OperationOutcome'):
        super(OperationOutcome, self).exportAttributes(outfile, level, already_processed, namespace_, name_='OperationOutcome')
    def exportChildren(self, outfile, level, namespace_='', name_='OperationOutcome', fromsubclass_=False, pretty_print=True):
        super(OperationOutcome, self).exportChildren(outfile, level, namespace_, name_, True, pretty_print=pretty_print)
        if pretty_print:
            eol_ = '\n'
        else:
            eol_ = ''
        for issue_ in self.issue:
            issue_.export(outfile, level, namespace_, name_='issue', pretty_print=pretty_print)
    def to_etree(self, parent_element=None, name_='OperationOutcome', mapping_=None):
        element = super(OperationOutcome, self).to_etree(parent_element, name_, mapping_)
        for issue_ in self.issue:
            issue_.to_etree(element, name_='issue', mapping_=mapping_)
        if mapping_ is not None:
            mapping_[self] = element
        return element
    def build(self, node):
        already_processed = set()
        self.buildAttributes(node, node.attrib, already_processed)
        for child in node:
            nodeName_ = Tag_pattern_.match(child.tag).groups()[-1]
            self.buildChildren(child, node, nodeName_)
        return self
    def buildAttributes(self, node, attrs, already_processed):
        super(OperationOutcome, self).buildAttributes(node, attrs, already_processed)
    def buildChildren(self, child_, node, nodeName_, fromsubclass_=False):
        if nodeName_ == 'issue':
            obj_ = OperationOutcome_Issue.factory()
            obj_.build(child_)
            self.issue.append(obj_)
            obj_.original_tagname_ = 'issue'
        super(OperationOutcome, self).buildChildren(child_, node, nodeName_, True)
# end class OperationOutcome


class OperationOutcome_Issue(BackboneElement):
    """A collection of error, warning or information messages that result
    from a system action."""
    subclass = None
    superclass = BackboneElement
    def __init__(self, id=None, extension=None, modifierExtension=None, severity=None, type_=None, details=None, location=None):
        self.original_tagname_ = None
        super(OperationOutcome_Issue, self).__init__(id, extension, modifierExtension, )
        self.severity = severity
        self.type_ = type_
        self.details = details
        if location is None:
            self.location = []
        else:
            self.location = location
    def factory(*args_, **kwargs_):
        if OperationOutcome_Issue.subclass:
            return OperationOutcome_Issue.subclass(*args_, **kwargs_)
        else:
            return OperationOutcome_Issue(*args_, **kwargs_)
    factory = staticmethod(factory)
    def get_severity(self): return self.severity
    def set_severity(self, severity): self.severity = severity
    def get_type(self): return self.type_
    def set_type(self, type_): self.type_ = type_
    def get_details(self): return self.details
    def set_details(self, details): self.details = details
    def get_location(self): return self.location
    def set_location(self, location): self.location = location
    def add_location(self, value): self.location.append(value)
    def insert_location(self, index, value): self.location[index] = value
    def hasContent_(self):
        if (
            self.severity is not None or
            self.type_ is not None or
            self.details is not None or
            self.location or
            super(OperationOutcome_Issue, self).hasContent_()
        ):
            return True
        else:
            return False
    def export(self, outfile, level, namespace_='', name_='OperationOutcome.Issue', namespacedef_='', pretty_print=True):
        if pretty_print:
            eol_ = '\n'
        else:
            eol_ = ''
        if self.original_tagname_ is not None:
            name_ = self.original_tagname_
        showIndent(outfile, level, pretty_print)
        outfile.write('<%s%s%s' % (namespace_, name_, namespacedef_ and ' ' + namespacedef_ or '', ))
        already_processed = set()
        self.exportAttributes(outfile, level, already_processed, namespace_, name_='OperationOutcome.Issue')
        if self.hasContent_():
            outfile.write('>%s' % (eol_, ))
            self.exportChildren(outfile, level + 1, namespace_='', name_='OperationOutcome.Issue', pretty_print=pretty_print)
            showIndent(outfile, level, pretty_print)
            outfile.write('</%s%s>%s' % (namespace_, name_, eol_))
        else:
            outfile.write('/>%s' % (eol_, ))
    def exportAttributes(self, outfile, level, already_processed, namespace_='', name_='OperationOutcome.Issue'):
        super(OperationOutcome_Issue, self).exportAttributes(outfile, level, already_processed, namespace_, name_='OperationOutcome.Issue')
    def exportChildren(self, outfile, level, namespace_='', name_='OperationOutcome.Issue', fromsubclass_=False, pretty_print=True):
        super(OperationOutcome_Issue, self).exportChildren(outfile, level, namespace_, name_, True, pretty_print=pretty_print)
        if pretty_print:
            eol_ = '\n'
        else:
            eol_ = ''
        if self.severity is not None:
            self.severity.export(outfile, level, namespace_, name_='severity', pretty_print=pretty_print)
        if self.type_ is not None:
            self.type_.export(outfile, level, namespace_, name_='type', pretty_print=pretty_print)
        if self.details is not None:
            self.details.export(outfile, level, namespace_, name_='details', pretty_print=pretty_print)
        for location_ in self.location:
            location_.export(outfile, level, namespace_, name_='location', pretty_print=pretty_print)
    def to_etree(self, parent_element=None, name_='OperationOutcome.Issue', mapping_=None):
        element = super(OperationOutcome_Issue, self).to_etree(parent_element, name_, mapping_)
        if self.severity is not None:
            severity_ = self.severity
            severity_.to_etree(element, name_='severity', mapping_=mapping_)
        if self.type_ is not None:
            type__ = self.type_
            type__.to_etree(element, name_='type', mapping_=mapping_)
        if self.details is not None:
            details_ = self.details
            details_.to_etree(element, name_='details', mapping_=mapping_)
        for location_ in self.location:
            location_.to_etree(element, name_='location', mapping_=mapping_)
        if mapping_ is not None:
            mapping_[self] = element
        return element
    def build(self, node):
        already_processed = set()
        self.buildAttributes(node, node.attrib, already_processed)
        for child in node:
            nodeName_ = Tag_pattern_.match(child.tag).groups()[-1]
            self.buildChildren(child, node, nodeName_)
        return self
    def buildAttributes(self, node, attrs, already_processed):
        super(OperationOutcome_Issue, self).buildAttributes(node, attrs, already_processed)
    def buildChildren(self, child_, node, nodeName_, fromsubclass_=False):
        if nodeName_ == 'severity':
            obj_ = IssueSeverity.factory()
            obj_.build(child_)
            self.severity = obj_
            obj_.original_tagname_ = 'severity'
        elif nodeName_ == 'type':
            obj_ = Coding.factory()
            obj_.build(child_)
            self.type_ = obj_
            obj_.original_tagname_ = 'type'
        elif nodeName_ == 'details':
            obj_ = string.factory()
            obj_.build(child_)
            self.details = obj_
            obj_.original_tagname_ = 'details'
        elif nodeName_ == 'location':
            obj_ = string.factory()
            obj_.build(child_)
            self.location.append(obj_)
            obj_.original_tagname_ = 'location'
        super(OperationOutcome_Issue, self).buildChildren(child_, node, nodeName_, True)
# end class OperationOutcome_Issue



# This file is part of Tryton.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
import urllib.parse
import urllib.request, urllib.parse, urllib.error
import string
import xml.dom.minidom
from pywebdav.lib import propfind
from pywebdav.lib.errors import DAV_NotFound, DAV_Error, DAV_Forbidden
from pywebdav.lib.utils import get_uriparentpath
from pywebdav.lib.constants import DAV_VERSION_1, DAV_VERSION_2
from trytond.modules.webdav.protocol import TrytonDAVInterface, LOCAL, \
        WebDAVAuthRequestHandler
from trytond.pool import Pool
from trytond.transaction import Transaction

domimpl = xml.dom.minidom.getDOMImplementation()

TrytonDAVInterface.PROPS['urn:ietf:params:xml:ns:caldav'] = (
    'calendar-description',
    'calendar-data',
    'calendar-home-set',
    'calendar-user-address-set',
    'schedule-inbox-URL',
    'schedule-outbox-URL',
    )
TrytonDAVInterface.PROPS['DAV:'] = tuple(list(TrytonDAVInterface.PROPS['DAV:'])
    + ['principal-collection-set'])
TrytonDAVInterface.M_NS['urn:ietf:params:xml:ns:caldav'] = '_get_caldav'
DAV_VERSION_1['version'] += ',calendar-access,calendar-schedule'
DAV_VERSION_2['version'] += ',calendar-access,calendar-schedule'

_mk_prop_response = propfind.PROPFIND.mk_prop_response


def mk_prop_response(self, uri, good_props, bad_props, doc):
    res = _mk_prop_response(self, uri, good_props, bad_props, doc)
    if not isinstance(uri, str):
        uri = uri.decode()
    parent_uri = get_uriparentpath(uri and uri.strip('/') or '')
    if not parent_uri:
        return res
    dbname, parent_uri = TrytonDAVInterface.get_dburi(parent_uri)
    if parent_uri in ('Calendars', 'Calendars/'):
        ad = doc.createElement('calendar')
        ad.setAttribute('xmlns', 'urn:ietf:params:xml:ns:caldav')
        # Disable groupdav attribute for iPhone
        # vc = doc.createElement('vevent-collection')
        # vc.setAttribute('xmlns', 'http://groupdav.org/')
        cols = res.getElementsByTagName('D:collection')
        if cols:
            cols[0].parentNode.appendChild(ad)
            # cols[0].parentNode.appendChild(vc)
    return res

propfind.PROPFIND.mk_prop_response = mk_prop_response


def _get_caldav_calendar_description(self, uri):
    dbname, dburi = self._get_dburi(uri)
    if not dbname:
        raise DAV_NotFound
    pool = Pool(Transaction().database.name)
    try:
        Collection = pool.get('webdav.collection')
    except KeyError:
        raise DAV_NotFound
    if not getattr(Collection, 'get_calendar_description', None):
        raise DAV_NotFound
    try:
        res = Collection.get_calendar_description(dburi, cache=LOCAL.cache)
    except DAV_Error as exception:
        self._log_exception(exception)
        raise
    except Exception as exception:
        self._log_exception(exception)
        raise DAV_Error(500)
    return res

TrytonDAVInterface._get_caldav_calendar_description = \
    _get_caldav_calendar_description


def _get_caldav_calendar_data(self, uri):
    dbname, dburi = self._get_dburi(uri)
    if not dbname:
        raise DAV_NotFound
    pool = Pool(Transaction().database.name)
    try:
        Collection = pool.get('webdav.collection')
    except KeyError:
        raise DAV_NotFound
    if not getattr(Collection, 'get_calendar_data', None):
        raise DAV_NotFound
    try:
        res = Collection.get_calendar_data(dburi, cache=LOCAL.cache)
    except DAV_Error as exception:
        self._log_exception(exception)
        raise
    except Exception as exception:
        self._log_exception(exception)
        raise DAV_Error(500)
    return res

TrytonDAVInterface._get_caldav_calendar_data = _get_caldav_calendar_data


def _get_caldav_calendar_home_set(self, uri):
    dbname, dburi = self._get_dburi(uri)
    if not dbname:
        raise DAV_NotFound
    pool = Pool(Transaction().database.name)
    try:
        Collection = pool.get('webdav.collection')
    except KeyError:
        raise DAV_NotFound
    if not getattr(Collection, 'get_calendar_home_set', None):
        raise DAV_NotFound
    try:
        res = Collection.get_calendar_home_set(dburi, cache=LOCAL.cache)
    except DAV_Error as exception:
        self._log_exception(exception)
        raise
    except Exception as exception:
        self._log_exception(exception)
        raise DAV_Error(500)
    uparts = list(urllib.parse.urlsplit(uri))
    uparts[2] = urllib.parse.quote(dbname + res)
    doc = domimpl.createDocument(None, 'href', None)
    href = doc.documentElement
    href.tagName = 'D:href'
    # iPhone doesn't handle "http" in href
    # huri = doc.createTextNode(urlparse.urlunsplit(uparts))
    huri = doc.createTextNode(urllib.parse.quote('/' + dbname + res))
    href.appendChild(huri)
    return href

TrytonDAVInterface._get_caldav_calendar_home_set = \
    _get_caldav_calendar_home_set


def _get_caldav_calendar_user_address_set(self, uri):
    dbname, dburi = self._get_dburi(uri)
    if not dbname:
        raise DAV_NotFound
    pool = Pool(Transaction().database.name)
    try:
        Collection = pool.get('webdav.collection')
    except KeyError:
        raise DAV_NotFound
    if not getattr(Collection, 'get_calendar_user_address_set', None):
        raise DAV_NotFound
    try:
        res = Collection.get_calendar_user_address_set(
            dburi, cache=LOCAL.cache)
    except DAV_Error as exception:
        self._log_exception(exception)
        raise
    except Exception as exception:
        self._log_exception(exception)
        raise DAV_Error(500)
    doc = domimpl.createDocument(None, 'href', None)
    href = doc.documentElement
    href.tagName = 'D:href'
    huri = doc.createTextNode('MAILTO:' + res)
    href.appendChild(huri)
    return href

TrytonDAVInterface._get_caldav_calendar_user_address_set = \
    _get_caldav_calendar_user_address_set


def _get_caldav_schedule_inbox_URL(self, uri):
    dbname, dburi = self._get_dburi(uri)
    if not dbname:
        raise DAV_NotFound
    pool = Pool(Transaction().database.name)
    try:
        Collection = pool.get('webdav.collection')
    except KeyError:
        raise DAV_NotFound
    if not getattr(Collection, 'get_schedule_inbox_URL', None):
        raise DAV_NotFound
    try:
        res = Collection.get_schedule_inbox_URL(dburi, cache=LOCAL.cache)
    except DAV_Error as exception:
        self._log_exception(exception)
        raise
    except Exception as exception:
        self._log_exception(exception)
        raise DAV_Error(500)
    uparts = list(urllib.parse.urlsplit(uri))
    uparts[2] = urllib.parse.quote(dbname + res)
    doc = domimpl.createDocument(None, 'href', None)
    href = doc.documentElement
    href.tagName = 'D:href'
    huri = doc.createTextNode(urllib.parse.urlunsplit(uparts))
    href.appendChild(huri)
    return href

TrytonDAVInterface._get_caldav_schedule_inbox_URL = \
    _get_caldav_schedule_inbox_URL


def _get_caldav_schedule_outbox_URL(self, uri):
    dbname, dburi = self._get_dburi(uri)
    if not dbname:
        raise DAV_NotFound
    pool = Pool(Transaction().database.name)
    try:
        Collection = pool.get('webdav.collection')
    except KeyError:
        raise DAV_NotFound
    if not getattr(Collection, 'get_schedule_outbox_URL', None):
        raise DAV_NotFound
    try:
        res = Collection.get_schedule_outbox_URL(dburi, cache=LOCAL.cache)
    except DAV_Error as exception:
        self._log_exception(exception)
        raise
    except Exception as exception:
        self._log_exception(exception)
        raise DAV_Error(500)
    uparts = list(urllib.parse.urlsplit(uri))
    uparts[2] = urllib.parse.quote(dbname + res)
    doc = domimpl.createDocument(None, 'href', None)
    href = doc.documentElement
    href.tagName = 'D:href'
    huri = doc.createTextNode(urllib.parse.urlunsplit(uparts))
    href.appendChild(huri)
    return href

TrytonDAVInterface._get_caldav_schedule_outbox_URL = \
    _get_caldav_schedule_outbox_URL

_prev_get_dav_principal_collection_set = hasattr(TrytonDAVInterface,
        '_get_dav_principal_collection_set') and \
                TrytonDAVInterface._get_dav_principal_collection_set or None


def _get_dav_principal_collection_set(self, uri):
    dbname, dburi = self._get_dburi(uri)
    if dburi.startswith('Calendars'):
        uparts = list(urllib.parse.urlsplit(uri.decode()))
        uparts[2] = urllib.parse.quote(dbname + '/Calendars/')
        doc = domimpl.createDocument(None, 'href', None)
        href = doc.documentElement
        href.tagName = 'D:href'
        huri = doc.createTextNode(urllib.parse.urlunsplit(uparts))
        href.appendChild(huri)
        return href
    if _prev_get_dav_principal_collection_set:
        return _prev_get_dav_principal_collection_set(self, uri)
    raise DAV_NotFound

TrytonDAVInterface._get_dav_principal_collection_set = \
    _get_dav_principal_collection_set


def _get_caldav_post(self, uri, body, contenttype=''):
    dbname, dburi = self._get_dburi(uri)
    if not dbname:
        raise DAV_Forbidden
    pool = Pool(Transaction().database.name)
    Calendar = pool.get('calendar.calendar')
    if not getattr(Calendar, 'post', None):
        raise DAV_NotFound
    try:
        res = Calendar.post(dburi, body)
    except DAV_Error as exception:
        self._log_exception(exception)
        raise
    except Exception as exception:
        self._log_exception(exception)
        raise DAV_Error(500)
    return res

TrytonDAVInterface._get_caldav_post = _get_caldav_post

_prev_do_POST = WebDAVAuthRequestHandler.do_POST


def do_POST(self):
    dc = self.IFACE_CLASS

    uri = urllib.parse.urljoin(self.get_baseuri(dc), self.path)
    uri = urllib.parse.unquote(uri)

    dbname, dburi = TrytonDAVInterface.get_dburi(uri)
    if dburi.startswith('Calendars'):
        # read the body
        body = None
        if 'Content-Length' in self.headers:
            l = self.headers['Content-Length']
            body = self.rfile.read(int(l))
        ct = None
        if 'Content-Type' in self.headers:
            ct = self.headers['Content-Type']

        try:
            DATA = '%s\n' % dc._get_caldav_post(uri, body, ct)
        except DAV_Error as exception:
            ec, _ = exception
            return self.send_status(ec)
        self.send_body(DATA, 200, 'OK', 'OK')
        return
    return _prev_do_POST(self)

WebDAVAuthRequestHandler.do_POST = do_POST





import pkg_resources

# try to get version from package (if installed)
try:
    package = pkg_resources.require('PyWebDAV3')[0]
    VERSION = package.version
except pkg_resources.DistributionNotFound:
    # Not running from installed version
    VERSION = "DEVELOPMENT"

# author hardcoded here
AUTHOR = 'Andrew Leech <andrew@alelec.net>, Simon Pamies <spamsch@gmail.com>'



from StringIO import StringIO
from operator import attrgetter
from .datastore import find_record
import server.fhir as supermod
from server.common import safe_attrgetter
from .health_mixin import ExportXMLMixin

# URL values
try:
    from flask import url_for
    RUN_FLASK=True
except:
    from .datastore import dumb_url_generate
    RUN_FLASK=False

class Immunization_Map:
    """Holds essential mappings and information for
        the Immunization class
    """

    root_search=[]

    resource_search_params={
            '_id': 'token',
            'date': 'date',
            'lot-number': 'string',
            'dose-sequence': 'number',
            'performer': 'reference',
            'vaccine-type': 'token',
            'subject': 'reference',
            '_language': None,
            'identifier': None,
            'location': None,
            'manufacturer': None,
            'reaction': None,
            'reaction-date': None,
            'reason': None,
            'refusal-reason': None,
            'refused': None,
            'requester': None}

    chain_map={ 'subject': 'Patient',
            'performer': 'Practitioner',
            'requester': 'Practitioner'}
            #'reaction': 'Observation', #TODO AdverseReaction
            #'location': 'Location', #TODO
            #'manufacturer': 'Organization'} #TODO

    search_mapping={
            '_id': ['id'],
            'date': ['date'],
            'subject': ['name'],
            'lot-number': ['lot.number'],
            'dose-sequence': ['dose'],
            'performer': ['healthprof'],
            'vaccine-type': ['vaccine.name.code'],
            'vaccine-type:text': ['vaccine.name.name']
        }

    url_prefixes={}
    model_mapping={'gnuhealth.vaccination':
            {
                'date': 'date',
                'expirationDate': 'lot.expiration_date',
                'lotNumber': 'lot.number',
                'performer': 'healthprof',
                'subject': 'name',
                'doseQuantity': 'amount',
                'site': 'admin_site',
                'route': 'admin_route',
                'vaccine-type': 'vaccine'
            }}


class health_Immunization(supermod.Immunization, Immunization_Map, ExportXMLMixin):

    def __init__(self, *args, **kwargs):
        rec = kwargs.pop('gnu_record', None)
        super(health_Immunization, self).__init__(*args, **kwargs)
        if rec:
            self.set_gnu_immunization(rec)

    def set_gnu_immunization(self, immunization):
        """Set immunization"""

        self.immunization = immunization
        self.model_type = self.immunization.__name__

        # Only certain models
        if self.model_type not in self.model_mapping:
            raise ValueError('Not a valid model')

        self.map = self.model_mapping[self.model_type]

        self.__import_from_gnu_immunization()
        self.__set_feed_info()

    def __import_from_gnu_immunization(self):
        """Import the data from the Health model"""

        if self.immunization:
            self.set_date(safe_attrgetter(self.immunization,
                    self.map['date']))
            self.set_expirationDate(safe_attrgetter(self.immunization,
                    self.map['expirationDate']))
            self.set_subject(safe_attrgetter(self.immunization,
                    self.map['subject']))
            self.set_performer(safe_attrgetter(self.immunization,
                    self.map['performer']))
            self.set_lotNumber(safe_attrgetter(self.immunization,
                    self.map['lotNumber']))
            self.set_doseQuantity(safe_attrgetter(self.immunization,
                    self.map['doseQuantity']))
            self.set_route(safe_attrgetter(self.immunization,
                    self.map['route']))
            self.set_site(safe_attrgetter(self.immunization,
                    self.map['site']))
            self.set_vaccineType(safe_attrgetter(self.immunization,
                    self.map['vaccine-type']))
            self.set_reported(reported=False)
            self.set_refusedIndicator(refusedIndicator=False)

    def __set_feed_info(self):
        """Set the feed-relevant data"""

        if self.immunization:
            if RUN_FLASK:
                uri = url_for('im_record',
                            log_id=self.immunization.id,
                            _external=True)
            else:
                uri = dumb_url_generate(['Immunization', self.immunization.id])
            self.feed={'id': uri,
                        'published': self.immunization.create_date,
                        'updated': self.immunization.write_date or \
                                self.immunization.create_date,
                        'title': '{} for {}'.format(
                            self.immunization.vaccine.name.name,
                            self.immunization.name.rec_name)
                        }

    def set_date(self, date):
        """Extends superclass for convenience

        Keyword arguments:
        date -- date (datetime object)
        """

        if date is not None:
            d = supermod.dateTime(value=date.strftime("%Y/%m/%d"))
            super(health_Immunization, self).set_date(d)

    def set_subject(self, subject):
        """Extends superclass for convenience

        Keyword arguments:
        subject -- patient (Health model)
        """

        if subject:
            if RUN_FLASK:
                uri = url_for('pat_record', log_id=subject.id)
            else:
                uri = dumb_url_generate(['Patient', subject.id])
            display = subject.rec_name
            ref=supermod.ResourceReference()
            ref.display = supermod.string(value=display)
            ref.reference = supermod.string(value=uri)
            super(health_Immunization, self).set_subject(ref)

    def set_performer(self, performer):
        """Extends superclass for convenience

        Keyword arguments:
        performer -- practitioner (Health model)
        """

        if performer:
            if RUN_FLASK:
                uri = url_for('hp_record', log_id=performer.id)
            else:
                uri = dumb_url_generate(['Practitioner', performer.id])
            display = performer.rec_name
            ref=supermod.ResourceReference()
            ref.display = supermod.string(value=display)
            ref.reference = supermod.string(value=uri)
            super(health_Immunization, self).set_performer(ref)

    def set_lotNumber(self, lotNumber):
        """Extends superclass for convenience

        Keyword arguments:
        lotNumber -- lot id (string, number, etc.)
        """

        if lotNumber is not None:
            s = supermod.string(value=str(lotNumber))
            super(health_Immunization, self).set_lotNumber(s)

    def set_expirationDate(self, expirationDate):
        """Extends superclass for convenince

        Keyword arguments:
        expirationDate -- date (datetime object)
        """

        if expirationDate is not None:
            d = supermod.date(value=expirationDate.strftime("%Y/%m/%d"))
            super(health_Immunization, self).set_expirationDate(d)

    def set_doseQuantity(self, doseQuantity):
        """Extends superclass for convenience

        Keyword arguments:
        doseQuantity -- amount in mL
        """

        if doseQuantity is not None:
            v = supermod.decimal(value=str(float(doseQuantity)))
            u = supermod.string(value='mL')
            s = supermod.uri(value='http://snomed.info/sct')
            c = supermod.code(value='258773002')
            dq = supermod.Quantity(value=v, units=u, system=s, code=c)
            super(health_Immunization, self).set_doseQuantity(dq)

    def set_refusedIndicator(self, refusedIndicator=False):
        """Extends superclass for convenience

        Keyword argument:
        refusedIndicator -- refused? True/False (default: False)
        """

        b = 'false'
        if refusedIndicator:
            b = 'true'
        ri = supermod.boolean(value=b)
        super(health_Immunization, self).set_refusedIndicator(ri)

    def set_reported(self, reported=False):
        """Extends superclass for convenience

        Keyword argument:
        reported -- self reported? True/False (default: False)
        """

        b = 'false'
        if reported:
            b = 'true'
        r = supermod.boolean(value=b)
        super(health_Immunization, self).set_reported(r)

    def set_route(self, route):
        """Extends superclass for convenience

        Keyword argument:
        route -- how vaccine entered body
        """

        from server.fhir.value_sets import immunizationRoute
        if route:
            ir=[i for i in immunizationRoute.contents if i['code'] == route.upper()]
            if ir:
                cc=supermod.CodeableConcept()
                c = supermod.Coding()
                c.display = cc.text = supermod.string(value=ir[0]['display'])
                c.code = supermod.code(value=ir[0]['code'])
                cc.coding=[c]
                super(health_Immunization, self).set_route(cc)

    def set_site(self, site):
        """Extends superclass for convenience

        Keyword arguments:
        site -- site code where the vaccine was administered
        """

        from server.fhir.value_sets import immunizationSite
        if site:
            m=[i for i in immunizationSite.contents if i['code'] == site.upper()]
            if m:
                cc=supermod.CodeableConcept()
                c = supermod.Coding()
                c.display = cc.text = supermod.string(value=m[0]['display'])
                c.code = supermod.code(value=m[0]['code'])
                cc.coding=[c]
                super(health_Immunization, self).set_site(cc)

    def set_vaccineType(self, vaccineType):
        """Extends superclass for convenience

        Keyword arguments:
        vaccineType -- the vaccine (Health model)
        """
        #TODO Need better coding, much better!

        if vaccineType:
            cc = supermod.CodeableConcept()
            c = supermod.Coding()
            c.display = cc.text = supermod.string(value=vaccineType.name.name)
            if vaccineType.name.code:
                c.code = supermod.code(value=vaccineType.name.code)
                cc.coding=[c]
            super(health_Immunization, self).set_vaccineType(cc)

supermod.Immunization.subclass=health_Immunization



# -*- coding: utf-8 -*-
##############################################################################
#
#    GNU Health: The Free Health and Hospital Information System
#    Copyright (C) 2008-2017 Luis Falcon <lfalcon@gnusolidario.org>
#    Copyright (C) 2011-2017 GNU Solidario <health@gnusolidario.org>
#
#
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################
from datetime import datetime
from trytond.model import ModelView, fields
from trytond.wizard import Wizard, StateTransition, StateView, Button
from trytond.transaction import Transaction
from trytond.pool import Pool


__all__ = ['CreateBedTransferInit', 'CreateBedTransfer']


class CreateBedTransferInit(ModelView):
    'Create Bed Transfer Init'
    __name__ = 'gnuhealth.bed.transfer.init'
    newbed = fields.Many2One('gnuhealth.hospital.bed', 'New Bed',
        required=True, select=True)
    reason = fields.Char('Reason', required=True)

    orig_bed_state = fields.Selection((
        (None,''),
        ('free', 'Free'),
        ('reserved', 'Reserved'),
        ('occupied', 'Occupied'),
        ('to_clean', 'Needs cleaning'),
        ('na', 'Not available'),
        ), 'Bed of origin Status', sort=False, required=True)

    @staticmethod
    def default_orig_bed_state():
        return 'to_clean'

class CreateBedTransfer(Wizard):
    'Create Bed Transfer'
    __name__ = 'gnuhealth.bed.transfer.create'


    start = StateView('gnuhealth.bed.transfer.init',
        'health_inpatient.view_patient_bed_transfer', [
            Button('Cancel', 'end', 'tryton-cancel'),
            Button('Transfer Patient', 'create_bed_transfer', 'tryton-ok',
                True),
            ])
    create_bed_transfer = StateTransition()


    @classmethod
    def __setup__(cls):
        super(CreateBedTransfer, cls).__setup__()
        cls._error_messages.update({
            'bed_unavailable': 'Destination bed is unavailable',
            'choose_one': 'You have chosen more than 1 records. Please choose only one'})

    def transition_create_bed_transfer(self):
        inpatient_registrations = Pool().get('gnuhealth.inpatient.registration')
        bed = Pool().get('gnuhealth.hospital.bed')

        registrations = inpatient_registrations.browse(Transaction().context.get(
            'active_ids'))

        # Don't allow mass changes. Work on a single record
        if len(registrations) > 1 :
            self.raise_user_error('choose_one')

        registration = registrations[0]
        current_bed = registration.bed
        destination_bed = self.start.newbed
        reason = self.start.reason
        orig_bed_state = self.start.orig_bed_state


        # Check that the new bed is free
        if (destination_bed.state == 'free'):

            # Update bed status with the one given in the transfer
            bed.write([current_bed], {'state': orig_bed_state})

            # Set as occupied the new bed
            bed.write([destination_bed], {'state': 'occupied'})
            # Update the hospitalization record
            hospitalization_info = {}

            hospitalization_info['bed'] = destination_bed

            # Update the hospitalization data
            transfers = []
            transfers.append(('create', [{
                            'transfer_date' : datetime.now(),
                            'bed_from' : current_bed,
                            'bed_to' : destination_bed,
                            'reason': reason,
                        }]))
            hospitalization_info['bed_transfers'] = transfers

            inpatient_registrations.write([registration],hospitalization_info)



        else:
            self.raise_user_error('bed_unavailable')

        return 'end'




from .base_classes import *
from .support_functions import *


class Practitioner(Resource):
    """A person who is directly or indirectly involved in the provisioning
    of healthcare.If the element is present, it must have either a
    @value, an @id, or extensions"""
    subclass = None
    superclass = Resource
    def __init__(self, id=None, extension=None, modifierExtension=None, language=None, text=None, contained=None, identifier=None, name=None, telecom=None, address=None, gender=None, birthDate=None, photo=None, organization=None, role=None, specialty=None, period=None, location=None, qualification=None, communication=None):
        self.original_tagname_ = None
        super(Practitioner, self).__init__(id, extension, modifierExtension, language, text, contained, )
        if identifier is None:
            self.identifier = []
        else:
            self.identifier = identifier
        self.name = name
        if telecom is None:
            self.telecom = []
        else:
            self.telecom = telecom
        self.address = address
        self.gender = gender
        self.birthDate = birthDate
        if photo is None:
            self.photo = []
        else:
            self.photo = photo
        self.organization = organization
        if role is None:
            self.role = []
        else:
            self.role = role
        if specialty is None:
            self.specialty = []
        else:
            self.specialty = specialty
        self.period = period
        if location is None:
            self.location = []
        else:
            self.location = location
        if qualification is None:
            self.qualification = []
        else:
            self.qualification = qualification
        if communication is None:
            self.communication = []
        else:
            self.communication = communication
    def factory(*args_, **kwargs_):
        if Practitioner.subclass:
            return Practitioner.subclass(*args_, **kwargs_)
        else:
            return Practitioner(*args_, **kwargs_)
    factory = staticmethod(factory)
    def get_identifier(self): return self.identifier
    def set_identifier(self, identifier): self.identifier = identifier
    def add_identifier(self, value): self.identifier.append(value)
    def insert_identifier(self, index, value): self.identifier[index] = value
    def get_name(self): return self.name
    def set_name(self, name): self.name = name
    def get_telecom(self): return self.telecom
    def set_telecom(self, telecom): self.telecom = telecom
    def add_telecom(self, value): self.telecom.append(value)
    def insert_telecom(self, index, value): self.telecom[index] = value
    def get_address(self): return self.address
    def set_address(self, address): self.address = address
    def get_gender(self): return self.gender
    def set_gender(self, gender): self.gender = gender
    def get_birthDate(self): return self.birthDate
    def set_birthDate(self, birthDate): self.birthDate = birthDate
    def get_photo(self): return self.photo
    def set_photo(self, photo): self.photo = photo
    def add_photo(self, value): self.photo.append(value)
    def insert_photo(self, index, value): self.photo[index] = value
    def get_organization(self): return self.organization
    def set_organization(self, organization): self.organization = organization
    def get_role(self): return self.role
    def set_role(self, role): self.role = role
    def add_role(self, value): self.role.append(value)
    def insert_role(self, index, value): self.role[index] = value
    def get_specialty(self): return self.specialty
    def set_specialty(self, specialty): self.specialty = specialty
    def add_specialty(self, value): self.specialty.append(value)
    def insert_specialty(self, index, value): self.specialty[index] = value
    def get_period(self): return self.period
    def set_period(self, period): self.period = period
    def get_location(self): return self.location
    def set_location(self, location): self.location = location
    def add_location(self, value): self.location.append(value)
    def insert_location(self, index, value): self.location[index] = value
    def get_qualification(self): return self.qualification
    def set_qualification(self, qualification): self.qualification = qualification
    def add_qualification(self, value): self.qualification.append(value)
    def insert_qualification(self, index, value): self.qualification[index] = value
    def get_communication(self): return self.communication
    def set_communication(self, communication): self.communication = communication
    def add_communication(self, value): self.communication.append(value)
    def insert_communication(self, index, value): self.communication[index] = value
    def hasContent_(self):
        if (
            self.identifier or
            self.name is not None or
            self.telecom or
            self.address is not None or
            self.gender is not None or
            self.birthDate is not None or
            self.photo or
            self.organization is not None or
            self.role or
            self.specialty or
            self.period is not None or
            self.location or
            self.qualification or
            self.communication or
            super(Practitioner, self).hasContent_()
        ):
            return True
        else:
            return False
    def export(self, outfile, level, namespace_='', name_='Practitioner', namespacedef_='', pretty_print=True):
        if pretty_print:
            eol_ = '\n'
        else:
            eol_ = ''
        if self.original_tagname_ is not None:
            name_ = self.original_tagname_
        showIndent(outfile, level, pretty_print)
        outfile.write('<%s%s%s' % (namespace_, name_, namespacedef_ and ' ' + namespacedef_ or '', ))
        already_processed = set()
        self.exportAttributes(outfile, level, already_processed, namespace_, name_='Practitioner')
        if self.hasContent_():
            outfile.write('>%s' % (eol_, ))
            self.exportChildren(outfile, level + 1, namespace_='', name_='Practitioner', pretty_print=pretty_print)
            showIndent(outfile, level, pretty_print)
            outfile.write('</%s%s>%s' % (namespace_, name_, eol_))
        else:
            outfile.write('/>%s' % (eol_, ))
    def exportAttributes(self, outfile, level, already_processed, namespace_='', name_='Practitioner'):
        super(Practitioner, self).exportAttributes(outfile, level, already_processed, namespace_, name_='Practitioner')
    def exportChildren(self, outfile, level, namespace_='', name_='Practitioner', fromsubclass_=False, pretty_print=True):
        super(Practitioner, self).exportChildren(outfile, level, namespace_, name_, True, pretty_print=pretty_print)
        if pretty_print:
            eol_ = '\n'
        else:
            eol_ = ''
        for identifier_ in self.identifier:
            identifier_.export(outfile, level, namespace_, name_='identifier', pretty_print=pretty_print)
        if self.name is not None:
            self.name.export(outfile, level, namespace_, name_='name', pretty_print=pretty_print)
        for telecom_ in self.telecom:
            telecom_.export(outfile, level, namespace_, name_='telecom', pretty_print=pretty_print)
        if self.address is not None:
            self.address.export(outfile, level, namespace_, name_='address', pretty_print=pretty_print)
        if self.gender is not None:
            self.gender.export(outfile, level, namespace_, name_='gender', pretty_print=pretty_print)
        if self.birthDate is not None:
            self.birthDate.export(outfile, level, namespace_, name_='birthDate', pretty_print=pretty_print)
        for photo_ in self.photo:
            photo_.export(outfile, level, namespace_, name_='photo', pretty_print=pretty_print)
        if self.organization is not None:
            self.organization.export(outfile, level, namespace_, name_='organization', pretty_print=pretty_print)
        for role_ in self.role:
            role_.export(outfile, level, namespace_, name_='role', pretty_print=pretty_print)
        for specialty_ in self.specialty:
            specialty_.export(outfile, level, namespace_, name_='specialty', pretty_print=pretty_print)
        if self.period is not None:
            self.period.export(outfile, level, namespace_, name_='period', pretty_print=pretty_print)
        for location_ in self.location:
            location_.export(outfile, level, namespace_, name_='location', pretty_print=pretty_print)
        for qualification_ in self.qualification:
            qualification_.export(outfile, level, namespace_, name_='qualification', pretty_print=pretty_print)
        for communication_ in self.communication:
            communication_.export(outfile, level, namespace_, name_='communication', pretty_print=pretty_print)
    def to_etree(self, parent_element=None, name_='Practitioner', mapping_=None):
        element = super(Practitioner, self).to_etree(parent_element, name_, mapping_)
        for identifier_ in self.identifier:
            identifier_.to_etree(element, name_='identifier', mapping_=mapping_)
        if self.name is not None:
            name_ = self.name
            name_.to_etree(element, name_='name', mapping_=mapping_)
        for telecom_ in self.telecom:
            telecom_.to_etree(element, name_='telecom', mapping_=mapping_)
        if self.address is not None:
            address_ = self.address
            address_.to_etree(element, name_='address', mapping_=mapping_)
        if self.gender is not None:
            gender_ = self.gender
            gender_.to_etree(element, name_='gender', mapping_=mapping_)
        if self.birthDate is not None:
            birthDate_ = self.birthDate
            birthDate_.to_etree(element, name_='birthDate', mapping_=mapping_)
        for photo_ in self.photo:
            photo_.to_etree(element, name_='photo', mapping_=mapping_)
        if self.organization is not None:
            organization_ = self.organization
            organization_.to_etree(element, name_='organization', mapping_=mapping_)
        for role_ in self.role:
            role_.to_etree(element, name_='role', mapping_=mapping_)
        for specialty_ in self.specialty:
            specialty_.to_etree(element, name_='specialty', mapping_=mapping_)
        if self.period is not None:
            period_ = self.period
            period_.to_etree(element, name_='period', mapping_=mapping_)
        for location_ in self.location:
            location_.to_etree(element, name_='location', mapping_=mapping_)
        for qualification_ in self.qualification:
            qualification_.to_etree(element, name_='qualification', mapping_=mapping_)
        for communication_ in self.communication:
            communication_.to_etree(element, name_='communication', mapping_=mapping_)
        if mapping_ is not None:
            mapping_[self] = element
        return element
    def build(self, node):
        already_processed = set()
        self.buildAttributes(node, node.attrib, already_processed)
        for child in node:
            nodeName_ = Tag_pattern_.match(child.tag).groups()[-1]
            self.buildChildren(child, node, nodeName_)
        return self
    def buildAttributes(self, node, attrs, already_processed):
        super(Practitioner, self).buildAttributes(node, attrs, already_processed)
    def buildChildren(self, child_, node, nodeName_, fromsubclass_=False):
        if nodeName_ == 'identifier':
            obj_ = Identifier.factory()
            obj_.build(child_)
            self.identifier.append(obj_)
            obj_.original_tagname_ = 'identifier'
        elif nodeName_ == 'name':
            obj_ = HumanName.factory()
            obj_.build(child_)
            self.name = obj_
            obj_.original_tagname_ = 'name'
        elif nodeName_ == 'telecom':
            obj_ = Contact.factory()
            obj_.build(child_)
            self.telecom.append(obj_)
            obj_.original_tagname_ = 'telecom'
        elif nodeName_ == 'address':
            obj_ = Address.factory()
            obj_.build(child_)
            self.address = obj_
            obj_.original_tagname_ = 'address'
        elif nodeName_ == 'gender':
            obj_ = CodeableConcept.factory()
            obj_.build(child_)
            self.gender = obj_
            obj_.original_tagname_ = 'gender'
        elif nodeName_ == 'birthDate':
            obj_ = dateTime.factory()
            obj_.build(child_)
            self.birthDate = obj_
            obj_.original_tagname_ = 'birthDate'
        elif nodeName_ == 'photo':
            obj_ = Attachment.factory()
            obj_.build(child_)
            self.photo.append(obj_)
            obj_.original_tagname_ = 'photo'
        elif nodeName_ == 'organization':
            obj_ = ResourceReference.factory()
            obj_.build(child_)
            self.organization = obj_
            obj_.original_tagname_ = 'organization'
        elif nodeName_ == 'role':
            obj_ = CodeableConcept.factory()
            obj_.build(child_)
            self.role.append(obj_)
            obj_.original_tagname_ = 'role'
        elif nodeName_ == 'specialty':
            obj_ = CodeableConcept.factory()
            obj_.build(child_)
            self.specialty.append(obj_)
            obj_.original_tagname_ = 'specialty'
        elif nodeName_ == 'period':
            obj_ = Period.factory()
            obj_.build(child_)
            self.period = obj_
            obj_.original_tagname_ = 'period'
        elif nodeName_ == 'location':
            obj_ = ResourceReference.factory()
            obj_.build(child_)
            self.location.append(obj_)
            obj_.original_tagname_ = 'location'
        elif nodeName_ == 'qualification':
            obj_ = Practitioner_Qualification.factory()
            obj_.build(child_)
            self.qualification.append(obj_)
            obj_.original_tagname_ = 'qualification'
        elif nodeName_ == 'communication':
            obj_ = CodeableConcept.factory()
            obj_.build(child_)
            self.communication.append(obj_)
            obj_.original_tagname_ = 'communication'
        super(Practitioner, self).buildChildren(child_, node, nodeName_, True)
# end class Practitioner


class Practitioner_Qualification(BackboneElement):
    """A person who is directly or indirectly involved in the provisioning
    of healthcare."""
    subclass = None
    superclass = BackboneElement
    def __init__(self, id=None, extension=None, modifierExtension=None, code=None, period=None, issuer=None):
        self.original_tagname_ = None
        super(Practitioner_Qualification, self).__init__(id, extension, modifierExtension, )
        self.code = code
        self.period = period
        self.issuer = issuer
    def factory(*args_, **kwargs_):
        if Practitioner_Qualification.subclass:
            return Practitioner_Qualification.subclass(*args_, **kwargs_)
        else:
            return Practitioner_Qualification(*args_, **kwargs_)
    factory = staticmethod(factory)
    def get_code(self): return self.code
    def set_code(self, code): self.code = code
    def get_period(self): return self.period
    def set_period(self, period): self.period = period
    def get_issuer(self): return self.issuer
    def set_issuer(self, issuer): self.issuer = issuer
    def hasContent_(self):
        if (
            self.code is not None or
            self.period is not None or
            self.issuer is not None or
            super(Practitioner_Qualification, self).hasContent_()
        ):
            return True
        else:
            return False
    def export(self, outfile, level, namespace_='', name_='Practitioner.Qualification', namespacedef_='', pretty_print=True):
        if pretty_print:
            eol_ = '\n'
        else:
            eol_ = ''
        if self.original_tagname_ is not None:
            name_ = self.original_tagname_
        showIndent(outfile, level, pretty_print)
        outfile.write('<%s%s%s' % (namespace_, name_, namespacedef_ and ' ' + namespacedef_ or '', ))
        already_processed = set()
        self.exportAttributes(outfile, level, already_processed, namespace_, name_='Practitioner.Qualification')
        if self.hasContent_():
            outfile.write('>%s' % (eol_, ))
            self.exportChildren(outfile, level + 1, namespace_='', name_='Practitioner.Qualification', pretty_print=pretty_print)
            showIndent(outfile, level, pretty_print)
            outfile.write('</%s%s>%s' % (namespace_, name_, eol_))
        else:
            outfile.write('/>%s' % (eol_, ))
    def exportAttributes(self, outfile, level, already_processed, namespace_='', name_='Practitioner.Qualification'):
        super(Practitioner_Qualification, self).exportAttributes(outfile, level, already_processed, namespace_, name_='Practitioner.Qualification')
    def exportChildren(self, outfile, level, namespace_='', name_='Practitioner.Qualification', fromsubclass_=False, pretty_print=True):
        super(Practitioner_Qualification, self).exportChildren(outfile, level, namespace_, name_, True, pretty_print=pretty_print)
        if pretty_print:
            eol_ = '\n'
        else:
            eol_ = ''
        if self.code is not None:
            self.code.export(outfile, level, namespace_, name_='code', pretty_print=pretty_print)
        if self.period is not None:
            self.period.export(outfile, level, namespace_, name_='period', pretty_print=pretty_print)
        if self.issuer is not None:
            self.issuer.export(outfile, level, namespace_, name_='issuer', pretty_print=pretty_print)
    def to_etree(self, parent_element=None, name_='Practitioner.Qualification', mapping_=None):
        element = super(Practitioner_Qualification, self).to_etree(parent_element, name_, mapping_)
        if self.code is not None:
            code_ = self.code
            code_.to_etree(element, name_='code', mapping_=mapping_)
        if self.period is not None:
            period_ = self.period
            period_.to_etree(element, name_='period', mapping_=mapping_)
        if self.issuer is not None:
            issuer_ = self.issuer
            issuer_.to_etree(element, name_='issuer', mapping_=mapping_)
        if mapping_ is not None:
            mapping_[self] = element
        return element
    def build(self, node):
        already_processed = set()
        self.buildAttributes(node, node.attrib, already_processed)
        for child in node:
            nodeName_ = Tag_pattern_.match(child.tag).groups()[-1]
            self.buildChildren(child, node, nodeName_)
        return self
    def buildAttributes(self, node, attrs, already_processed):
        super(Practitioner_Qualification, self).buildAttributes(node, attrs, already_processed)
    def buildChildren(self, child_, node, nodeName_, fromsubclass_=False):
        if nodeName_ == 'code':
            obj_ = CodeableConcept.factory()
            obj_.build(child_)
            self.code = obj_
            obj_.original_tagname_ = 'code'
        elif nodeName_ == 'period':
            obj_ = Period.factory()
            obj_.build(child_)
            self.period = obj_
            obj_.original_tagname_ = 'period'
        elif nodeName_ == 'issuer':
            obj_ = ResourceReference.factory()
            obj_.build(child_)
            self.issuer = obj_
            obj_.original_tagname_ = 'issuer'
        super(Practitioner_Qualification, self).buildChildren(child_, node, nodeName_, True)
# end class Practitioner_Qualification
